# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for DB constraints."""

from typing import Any
from unittest import mock

from django.db import IntegrityError
from django.db.backends.base.schema import BaseDatabaseSchemaEditor
from django.db.models import CheckConstraint, Q

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.db.constraints import JsonDataUniqueConstraint
from debusine.db.models import Collection, CollectionItem, User
from debusine.test.django import TestCase


class JsonDataUniqueConstraintTests(TestCase):
    """Tests for JsonDataUniqueConstraint class."""

    def test_eq_true(self) -> None:
        """Test equality for equal JsonDataUniqueConstraint."""
        self.assertEqual(
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="unique",
            ),
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="unique",
            ),
        )

    def test_eq_false(self) -> None:
        """Test equality for not-equal JsonDataUniqueConstraint."""
        self.assertNotEqual(
            JsonDataUniqueConstraint(
                fields=["data->>architecture", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="unique",
            ),
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="unique",
            ),
        )

        # Different condition
        self.assertNotEqual(
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=True),
                nulls_distinct=False,
                name="unique",
            ),
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="unique",
            ),
        )

        # Different nulls_distinct
        self.assertNotEqual(
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=True,
                name="unique",
            ),
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="unique",
            ),
        )

        # Different name
        self.assertNotEqual(
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="name_1",
            ),
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="name_2",
            ),
        )

        # Comparing with a Constraint that is not a JsonDataUniqueConstraint
        self.assertNotEqual(
            JsonDataUniqueConstraint(
                fields=["data->>codename", "parent_collection"],
                condition=Q(category__isnull=False),
                nulls_distinct=False,
                name="name_1",
            ),
            CheckConstraint(check=Q(), name="price"),
        )

    def assert_constraint_deconstruction(
        self,
        constraint: JsonDataUniqueConstraint,
        expected_kwargs: dict[str, Any],
    ) -> None:
        """Assert constraint.deconstruct return expected path, args, kwargs."""
        path, args, kwargs = constraint.deconstruct()
        self.assertEqual(
            path, "debusine.db.constraints.JsonDataUniqueConstraint"
        )
        self.assertEqual(args, ())
        self.assertEqual(kwargs, expected_kwargs)

    def test_deconstruction(self) -> None:
        """Test deconstruction."""
        fields = ["foo", "bar"]
        name = "unique_fields"
        condition = Q(name__eq="name")

        constraint = JsonDataUniqueConstraint(
            fields=fields, name=name, nulls_distinct=True, condition=condition
        )
        expected_kwargs = {
            "fields": tuple(fields),
            "name": "unique_fields",
            "nulls_distinct": True,
            "condition": condition,
        }

        self.assert_constraint_deconstruction(constraint, expected_kwargs)

    def test_deconstruction_no_condition_no_nulls_distinct(self) -> None:
        """Test deconstruction without condition and without nulls distinct."""
        fields = ["foo", "bar"]
        name = "unique_fields"

        constraint = JsonDataUniqueConstraint(
            fields=fields, name=name, nulls_distinct=None, condition=Q()
        )
        expected_kwargs = {
            "fields": tuple(fields),
            "name": "unique_fields",
        }

        self.assert_constraint_deconstruction(constraint, expected_kwargs)

    def test_database_constraint(self) -> None:
        """
        Test cannot create two objects with duplicated data.

        Use CollectionItem debian:environments constraint to exercise relevant
        code:
        -The constraint works
        -Applies only on the relevant condition

        Complementary tests assert the SQL created and other code paths
        not used by CollectionItem debian:environments constraint.
        """
        self.workspace = self.playground.create_workspace()
        self.artifact, _ = self.playground.create_artifact()

        self.user = User.objects.create(
            username="John", email="john@example.com"
        )

        collection_test = Collection.objects.create(
            name="test",
            category=CollectionCategory.TEST,
            workspace=self.workspace,
        )
        collection_environments = Collection.objects.create(
            name="debian",
            category=CollectionCategory.ENVIRONMENTS,
            workspace=self.workspace,
        )

        item_kwargs = {
            "child_type": CollectionItem.Types.ARTIFACT,
            "artifact": self.artifact,
            "created_by_user": self.user,
        }

        # Allowed by the condition: category does not have any unique constraint
        CollectionItem.objects.create(
            **item_kwargs,
            name="name1",
            category=ArtifactCategory.TEST,
            parent_collection=collection_test,
        )
        CollectionItem.objects.create(
            **item_kwargs,
            name="name2",
            category=ArtifactCategory.TEST,
            parent_collection=collection_test,
        )

        CollectionItem.objects.create(
            **item_kwargs,
            name="name3",
            category=ArtifactCategory.SYSTEM_TARBALL,
            parent_collection=collection_environments,
            data={"codename": "bookworm", "architecture": "amd64"},
        )

        # Having both a tarball and an image with the same properties is
        # allowed
        CollectionItem.objects.create(
            **item_kwargs,
            name="name5",
            category=ArtifactCategory.SYSTEM_IMAGE,
            parent_collection=collection_environments,
            data={"codename": "bookworm", "architecture": "amd64"},
        )

        # Not allowed: the category debian:environments does not allow
        # duplicated codename and architecture in the same collection
        msg_contents = "db_collectionitem_unique_debian_environments"
        with self.assertRaisesRegex(IntegrityError, msg_contents):
            CollectionItem.objects.create(
                **item_kwargs,
                name="name4",
                category=ArtifactCategory.SYSTEM_TARBALL,
                parent_collection=collection_environments,
                data={"codename": "bookworm", "architecture": "amd64"},
            )

    @staticmethod
    def run_create_sql(
        constraint: JsonDataUniqueConstraint, *, table_name: str
    ) -> str:
        """Run constraint.create_sql and return SQL."""
        schema_editor = mock.create_autospec(
            spec=BaseDatabaseSchemaEditor, instance=True
        )
        schema_editor.connection = mock.Mock()
        schema_editor.connection.vendor = "postgresql"
        schema_editor.connection.ops = mock.Mock()
        schema_editor.connection.ops.quote_name = mock.Mock(
            return_value=f'"{table_name}"'
        )

        sql = constraint.create_sql(CollectionItem, schema_editor)

        return str(sql)

    def test_create_sql(self) -> None:
        """
        Test create_sql method.

        Covers code path condition=None not covered by test_database_constraint.
        """
        fields = ["name", "category"]
        name = "index_name"
        table_name = "table"
        constraint = JsonDataUniqueConstraint(
            fields=fields,
            name=name,
            nulls_distinct=False,
            condition=None,
        )
        fields_str = ", ".join(fields)
        actual_sql = self.run_create_sql(constraint, table_name=table_name)
        expected_sql = (
            f'CREATE UNIQUE INDEX {name} ON "{table_name}" ({fields_str}) '
            f'NULLS NOT DISTINCT'
        )
        self.assertEqual(actual_sql, expected_sql)

    def test_create_sql_nulls_distinct_false(self) -> None:
        """Test create_sql adds NULLS NOT DISTINCT."""
        constraint = JsonDataUniqueConstraint(
            fields=["name", "data->>'variant'"],
            name="index_name",
            nulls_distinct=False,
            condition=None,
        )
        table_name = "table_name"
        actual_sql = self.run_create_sql(constraint, table_name=table_name)
        expected_sql = (
            f'CREATE UNIQUE INDEX index_name ON "{table_name}" '
            f"(name, (data->>'variant')) NULLS NOT DISTINCT"
        )
        self.assertEqual(actual_sql, expected_sql)

    def test_create_sql_nulls_distinct_true(self) -> None:
        """Test create_sql adds NULLS DISTINCT."""
        constraint = JsonDataUniqueConstraint(
            fields=["name", "data->>'variant'"],
            name="index_name",
            nulls_distinct=True,
            condition=None,
        )
        table_name = "table"
        actual_sql = self.run_create_sql(constraint, table_name=table_name)
        expected_sql = (
            f'CREATE UNIQUE INDEX index_name ON "{table_name}" '
            f"(name, (data->>'variant')) NULLS DISTINCT"
        )

        self.assertEqual(actual_sql, expected_sql)

    def test_create_sql_distinct_none(self) -> None:
        """Test create_sql does not add NULLS (NOT) DISTINCT."""
        constraint = JsonDataUniqueConstraint(
            fields=["name", "data->>'variant'"],
            name="index_name",
            nulls_distinct=None,
            condition=None,
        )
        table_name = "table"
        actual_sql = self.run_create_sql(constraint, table_name=table_name)
        expected_sql = (
            f'CREATE UNIQUE INDEX index_name ON "{table_name}" '
            "(name, (data->>'variant'))"
        )

        self.assertEqual(actual_sql, expected_sql)

    def test_remove_sql(self) -> None:
        """Test remove_sql drop the constraint index."""
        name = "index_name"
        constraint = JsonDataUniqueConstraint(
            fields=["not_used"],
            name=name,
            nulls_distinct=False,
            condition=None,
        )

        self.assertEqual(
            str(constraint.remove_sql(None, None)), f"DROP INDEX {name}"
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the Incus executor."""

import tarfile
from io import BytesIO
from pathlib import Path
from shutil import rmtree
from subprocess import CalledProcessError, CompletedProcess, DEVNULL, PIPE
from tempfile import NamedTemporaryFile, mkdtemp
from textwrap import dedent
from typing import Any, Literal, cast, override
from unittest.mock import ANY, MagicMock, call, create_autospec, patch

import yaml

from debusine.client.debusine import Debusine
from debusine.client.models import ArtifactResponse
from debusine.tasks import Noop
from debusine.tasks.executors.base import (
    ExecutorImageCategory,
    ExecutorInterface,
    ExecutorStatistics,
    ImageImportError,
    InstanceNotRunning,
    executor_class,
)
from debusine.tasks.executors.incus import (
    IncusExecutorMixin,
    IncusImageCache,
    IncusInstance,
    IncusLXCExecutor,
    IncusLXCImageCache,
    IncusVMExecutor,
    IncusVMImageCache,
    log,
    run_incus_cmd,
)
from debusine.tasks.models import BackendType
from debusine.tasks.tests.helpers import ExternalTaskHelperMixin
from debusine.test import TestCase


class RunIncusCmdTests(TestCase):
    """Unit tests for run_incus_cmd."""

    @patch("debusine.tasks.executors.incus.check_output")
    def test_run_incus_cmd(self, check_output: MagicMock) -> None:
        """Test that run_incus_cmd is a thin wrapper around check_output."""
        check_output.return_value = "output"
        output = run_incus_cmd("testing", "123")
        self.assertEqual(output, "output")
        check_output.assert_called_with(
            ["incus", "testing", "123"],
            text=True,
            encoding="utf-8",
            stderr=None,
        )

    @patch("debusine.tasks.executors.incus.check_output")
    def test_run_incus_cmd_suppress_stderr(
        self, check_output: MagicMock
    ) -> None:
        """Test that run_incus_cmd can suppress stderr."""
        check_output.return_value = "output"
        output = run_incus_cmd("testing", "123", suppress_stderr=True)
        self.assertEqual(output, "output")
        check_output.assert_called_with(
            ["incus", "testing", "123"],
            text=True,
            encoding="utf-8",
            stderr=DEVNULL,
        )


class IncusImageCacheCommonTests(ExternalTaskHelperMixin[Noop], TestCase):
    """Unit tests that apply to all the Incus ImageCache implementations."""

    image_cache_cls: type[IncusImageCache]
    artifact: ArtifactResponse
    artifact_base_name: str
    artifact_category: ExecutorImageCategory

    @override
    def setUp(self) -> None:
        """Mock the image handling for tests."""
        super().setUp()
        self.debusine_api = create_autospec(Debusine)
        self.image_cache = self.image_cache_cls(
            self.debusine_api, self.artifact_category, workspace="workspace"
        )
        self.image_cache.image_cache_path = Path("/tmp/not-used")

    def create_tar(
        self,
        path: Path | None = None,
        dirs: tuple[str, ...] = ("./", "./sbin"),
        files: tuple[str, ...] = ("./test-file", "./sbin/init"),
        special_files: bool = False,
    ) -> Path:
        """
        Create a test tar file, creating a temporary file if needed.

        The named dirs will be created at the beginning of the file.
        The named files will be created with dummy content.
        If special_files is set, a full range of file types will be included.
        """
        if path is None:
            path = Path(
                NamedTemporaryFile(  # noqa: SIM115
                    prefix="debusine-test-", suffix=".tar", delete=False
                ).name
            )
            cast(TestCase, self).addCleanup(path.unlink)

        mode: Literal["w", "w:xz"]
        if path.suffix == ".xz":
            mode = "w:xz"
        else:
            mode = "w"

        with tarfile.open(name=path, mode=mode) as tar:
            for dir_ in dirs:
                tarinfo = tarfile.TarInfo(name=dir_)
                tarinfo.type = tarfile.DIRTYPE
                tarinfo.mode = 0o755
                tar.addfile(tarinfo)
            for file in files:
                tarinfo = tarfile.TarInfo(name=file)
                tarinfo.size = 3
                tar.addfile(tarinfo, BytesIO(b"foo"))
            if special_files:
                tarinfo = tarfile.TarInfo(name="./special")
                tarinfo.type = tarfile.DIRTYPE
                tar.addfile(tarinfo)
                tarinfo = tarfile.TarInfo(name="./special/reg-file")
                tar.addfile(tarinfo, BytesIO(b"foo"))
                tarinfo = tarfile.TarInfo(name="./special/hard-link")
                tarinfo.type = tarfile.LNKTYPE
                tarinfo.linkname = "./special/reg-file"
                tar.addfile(tarinfo)
                tarinfo = tarfile.TarInfo(name="./special/sym-link")
                tarinfo.type = tarfile.SYMTYPE
                tarinfo.linkname = "/dev/null"
                tar.addfile(tarinfo)
                tarinfo = tarfile.TarInfo(name="./special/chr-dev")
                tarinfo.type = tarfile.CHRTYPE
                tarinfo.devmajor = 10
                tarinfo.devminor = 42
                tar.addfile(tarinfo)
                tarinfo = tarfile.TarInfo(name="./special/blk-dev")
                tarinfo.type = tarfile.BLKTYPE
                tarinfo.devmajor = 11
                tarinfo.devminor = 42
                tar.addfile(tarinfo)
                tarinfo = tarfile.TarInfo(name="./special/fifo-dev")
                tarinfo.type = tarfile.FIFOTYPE
                tar.addfile(tarinfo)

        return path

    @patch("debusine.tasks.executors.incus.run_incus_cmd")
    def test_image_in_cache_uncached(self, run_incus_cmd: MagicMock) -> None:
        """Test that image_in_cache fails to find a cached image."""
        run_incus_cmd.side_effect = CalledProcessError(
            cmd=["incus"], returncode=1
        )
        self.assertFalse(self.image_cache.image_in_cache(self.artifact))
        run_incus_cmd.assert_called_with(
            "image", "info", "artifact/42", suppress_stderr=True
        )

    @patch("debusine.tasks.executors.incus.run_incus_cmd")
    def test_image_in_cache_cached(self, run_incus_cmd: MagicMock) -> None:
        """Test that image_in_cache finds cached image."""
        run_incus_cmd.return_value = "Fingerprint: abc123\n"
        self.assertTrue(self.image_cache.image_in_cache(self.artifact))
        run_incus_cmd.assert_called_with(
            "image", "info", "artifact/42", suppress_stderr=True
        )

    @patch("debusine.tasks.executors.incus.run_incus_cmd")
    def test_process_downloaded_image(self, run_incus_cmd: MagicMock) -> None:
        """
        Test that process_downloaded_image calls build_incus_image and uploads.

        The real work is in build_incus_image(), so we mock everything and
        ensure that the expected calls are made.
        """
        built_path = Path("incus.tar.xz")
        with patch.object(
            self.image_cache,
            "build_incus_image",
            return_value=built_path,
        ) as build_incus_image:
            self.image_cache.process_downloaded_image(self.artifact)

        downloaded_path = self.image_cache.cache_artifact_image_path(
            42, self.artifact_base_name
        )
        build_incus_image.assert_called_with(
            downloaded_path, ANY, self.artifact
        )
        run_incus_cmd.assert_called_with(
            "image", "import", "incus.tar.xz", "--alias", "artifact/42"
        )

    def test_image_metadata(self) -> None:
        """Test that _image_metadata returns the correct metadata."""
        metadata = self.image_cache._image_metadata(self.artifact)

        expected = {
            "architecture": "x86_64",
            "creation_date": 1704067200,
            "profiles": ["debusine"],
            "properties": {
                "debian_architecture": "amd64",
                "debusine_artifact_id": 42,
                "description": (
                    f"Debusine Artifact #42: "
                    f"{self.artifact_category} for debian bookworm amd64"
                ),
                "os": "debian",
                "release": "bookworm",
            },
            "templates": {
                "/etc/hostname": {
                    "create_only": False,
                    "properties": {},
                    "template": "hostname.tpl",
                    "when": ["create", "copy"],
                },
                "/etc/hosts": {
                    "create_only": False,
                    "properties": {},
                    "template": "hosts.tpl",
                    "when": ["create", "copy"],
                },
            },
        }
        self.assertEqual(metadata, expected)

    def test_add_tar_directory(self) -> None:
        """Test that the add_tar_directory helper adds a directory."""
        buffer = BytesIO()
        with tarfile.TarFile(fileobj=buffer, mode="w") as tar:
            self.image_cache.add_tar_directory(tar, "foo")

        buffer.seek(0)

        with tarfile.TarFile(fileobj=buffer, mode="r") as tar:
            members = tar.getmembers()
            self.assertEqual(len(members), 1)
            member = members[0]
            self.assertEqual(member.name, "foo")
            self.assertTrue(member.isdir())
            self.assertEqual(member.mode, 0o755)
            self.assertEqual(member.uid, 0)
            self.assertEqual(member.gid, 0)

    def test_add_tar_file(self) -> None:
        """Test that the add_tar_file helper adds a file."""
        buffer = BytesIO()
        with tarfile.TarFile(fileobj=buffer, mode="w") as tar:
            self.image_cache.add_tar_file(tar, "foo", "bar")

        buffer.seek(0)

        with tarfile.TarFile(fileobj=buffer, mode="r") as tar:
            members = tar.getmembers()
            self.assertEqual(len(members), 1)
            member = members[0]
            self.assertEqual(member.name, "foo")
            self.assertTrue(member.isreg())
            self.assertEqual(member.mode, 0o644)
            self.assertEqual(member.uid, 0)
            self.assertEqual(member.gid, 0)
            member_file = tar.extractfile(member)
            assert member_file is not None
            self.assertEqual(member_file.read(), b"bar")

    def test_add_tar_symlink(self) -> None:
        """Test that the add_tar_symlink helper adds a symlink."""
        buffer = BytesIO()
        with tarfile.TarFile(fileobj=buffer, mode="w") as tar:
            self.image_cache.add_tar_symlink(tar, "foo", "bar")

        buffer.seek(0)

        with tarfile.TarFile(fileobj=buffer, mode="r") as tar:
            members = tar.getmembers()
            self.assertEqual(len(members), 1)
            member = members[0]
            self.assertEqual(member.name, "foo")
            self.assertTrue(member.issym())
            self.assertEqual(member.uid, 0)
            self.assertEqual(member.gid, 0)
            self.assertEqual(member.linkname, "bar")


class IncusLXCImageCacheTests(IncusImageCacheCommonTests):
    """Unit tests for IncusLXCImageCache."""

    image_cache_cls = IncusLXCImageCache
    artifact_base_name = "system.tar.xz"
    artifact_category = ExecutorImageCategory.TARBALL

    @override
    def setUp(self) -> None:
        """Mock the image handling for tests."""
        super().setUp()
        self.artifact = self.fake_system_tarball_artifact()

    def test_prepare_incus_image(self) -> None:
        """Test that prepare copies over the image and metadata."""
        source = self.create_tar(
            dirs=("./", "./etc", "./sbin"), special_files=True
        )
        dest_buf = BytesIO()
        with tarfile.TarFile(fileobj=dest_buf, mode="w") as dest_tar:
            self.image_cache.prepare_incus_image(
                source, dest_tar, self.artifact
            )

        dest_buf.seek(0)

        with tarfile.open(fileobj=dest_buf) as tar:
            names = tar.getnames()
            self.assertEqual(
                names,
                [
                    "metadata.yaml",
                    "templates",
                    "templates/hosts.tpl",
                    "templates/hostname.tpl",
                    "rootfs",
                    "rootfs/etc",
                    "rootfs/sbin",
                    "rootfs/test-file",
                    "rootfs/sbin/init",
                    "rootfs/special",
                    "rootfs/special/reg-file",
                    "rootfs/special/hard-link",
                    "rootfs/special/sym-link",
                    "rootfs/special/chr-dev",
                    "rootfs/special/blk-dev",
                    "rootfs/special/fifo-dev",
                    "rootfs/etc/systemd",
                    "rootfs/etc/systemd/system",
                    "rootfs/etc/systemd/system-generators",
                    "rootfs/etc/systemd/system/multi-user.target.wants",
                    (
                        "rootfs/etc/systemd/system/multi-user.target.wants/"
                        "systemd-networkd.service"
                    ),
                    (
                        "rootfs/etc/systemd/system/network-online.target.wants/"
                        "systemd-networkd-wait-online.service"
                    ),
                    "rootfs/etc/systemd/system-generators/lxc",
                ],
            )

            tarinfo = tar.getmember("rootfs/special/hard-link")
            self.assertEqual(tarinfo.linkname, "rootfs/special/reg-file")
            tarinfo = tar.getmember("rootfs/special/sym-link")
            self.assertEqual(tarinfo.linkname, "/dev/null")

    def test_build_incus_image(self) -> None:
        """Test that build_incus_image injects basic metadata into a new tar."""
        directory = self.create_temporary_directory()
        source_path = directory / "system.tar.xz"
        self.create_tar(source_path)
        workdir = directory / "workdir"
        workdir.mkdir()

        image = self.image_cache.build_incus_image(
            source_path, workdir, self.artifact
        )

        self.assertTrue(image.exists())
        with tarfile.open(image) as tar:
            names = set(tar.getnames())
            self.assertIn("metadata.yaml", names)
            metadata_file = tar.extractfile("metadata.yaml")
            assert metadata_file is not None
            metadata = yaml.safe_load(metadata_file)

        self.assertNotIn("test-file", names)
        self.assertEqual(
            metadata, self.image_cache._image_metadata(self.artifact)
        )

    def test_prepare_incus_image_requires_init(self) -> None:
        """Test that prepare only accepts bootable images."""
        source = self.create_tar(files=())
        dest_buf = BytesIO()
        with (
            tarfile.TarFile(fileobj=dest_buf, mode="w") as dest_tar,
            self.assertRaises(ImageImportError),
        ):
            self.image_cache.prepare_incus_image(
                source, dest_tar, self.artifact
            )

    def test_prepare_incus_image_accepts_usrmerged_init(self) -> None:
        """Test that prepare accepts a /usr/sbin/init."""
        source = self.create_tar(
            dirs=("./", "./usr", "./usr/sbin"),
            files=("./usr/sbin/init",),
        )
        dest_buf = BytesIO()
        with tarfile.TarFile(fileobj=dest_buf, mode="w") as dest_tar:
            self.image_cache.prepare_incus_image(
                source, dest_tar, self.artifact
            )


class IncusVMImageCacheTests(IncusImageCacheCommonTests):
    """Unit tests for IncusVMImageCache."""

    image_cache_cls = IncusVMImageCache
    artifact_base_name = "image.qcow2"
    artifact_category = ExecutorImageCategory.IMAGE

    @override
    def setUp(self) -> None:
        """Mock the image handling for tests."""
        super().setUp()
        self.artifact = self.fake_system_image_artifact()

    def test_prepare_qcow2_incus_image(self) -> None:
        """Test that prepare converts qcow2 images."""
        with NamedTemporaryFile(
            prefix="debusine-test-", suffix=".qcow2", delete=False
        ) as source:
            source.write(b"qcow!")
            source_path = Path(source.name)
            self.addCleanup(source_path.unlink)
        dest_buf = BytesIO()
        with tarfile.TarFile(fileobj=dest_buf, mode="w") as dest_tar:
            self.image_cache.prepare_incus_image(
                source_path, dest_tar, self.artifact
            )

        dest_buf.seek(0)

        with tarfile.open(fileobj=dest_buf) as tar:
            names = tar.getnames()
            image_file = tar.extractfile("rootfs.img")
            assert image_file is not None
            image = image_file.read()

        self.assertEqual(
            names,
            [
                "metadata.yaml",
                "templates",
                "templates/hosts.tpl",
                "templates/hostname.tpl",
                "rootfs.img",
            ],
        )
        self.assertEqual(image, b"qcow!")

    @patch("debusine.tasks.executors.incus.check_call")
    def test_prepare_raw_incus_image(self, check_call: MagicMock) -> None:
        """Test that prepare converts raw images."""
        self.artifact.data["image_format"] = "raw"
        self.artifact.data["filename"] = "image.tar.xz"

        directory = Path(mkdtemp(prefix="debusine-tests-"))
        self.addCleanup(rmtree, directory)
        source = self.create_tar(
            path=directory / "image.tar.xz",
            files=("img.raw",),
            dirs=(),
        )
        dest_buf = BytesIO()

        def fake_convert(args: list[str]) -> None:
            """Fake for subprocess.check_call."""
            assert args[:2] == ["qemu-img", "convert"]
            with Path(args[-1]).open("wb") as f:
                f.write(b"qcow!")

        check_call.side_effect = fake_convert

        with tarfile.TarFile(fileobj=dest_buf, mode="w") as dest_tar:
            self.image_cache.prepare_incus_image(
                source, dest_tar, self.artifact
            )

        extracted = directory / "img.raw"
        self.assertFalse(extracted.exists())

        check_call.assert_called_with(
            [
                "qemu-img",
                "convert",
                "-f",
                "raw",
                "-O",
                "qcow2",
                str(extracted),
                str(directory / "image.qcow2"),
            ]
        )

        dest_buf.seek(0)

        with tarfile.open(fileobj=dest_buf) as tar:
            names = tar.getnames()
            image_file = tar.extractfile("rootfs.img")
            assert image_file is not None
            image = image_file.read()
            metadata_file = tar.extractfile("metadata.yaml")
            assert metadata_file is not None
            metadata = yaml.safe_load(metadata_file)

        self.assertEqual(
            names,
            [
                "metadata.yaml",
                "templates",
                "templates/hosts.tpl",
                "templates/hostname.tpl",
                "rootfs.img",
            ],
        )
        self.assertEqual(image, b"qcow!")
        self.assertEqual(
            metadata, self.image_cache._image_metadata(self.artifact)
        )


class IncusExecutorCommonTests(ExternalTaskHelperMixin[Noop], TestCase):
    """Unit tests that apply to both Incus Executors."""

    executor_cls: type[ExecutorInterface]
    uses_images: bool

    @override
    def setUp(self) -> None:
        """Mock the Debusine API for tests."""
        super().setUp()
        self.debusine_api = MagicMock(spec=Debusine)
        self.image_artifact = self.mock_image_download(
            self.debusine_api, system_image=self.uses_images
        )

        run_incus_cmd_patcher = patch(
            "debusine.tasks.executors.incus.run_incus_cmd"
        )
        self.run_incus_cmd = run_incus_cmd_patcher.start()
        self.addCleanup(run_incus_cmd_patcher.stop)

        self.workdir = self.create_temporary_directory()
        self.executor = self.executor_cls(
            self.debusine_api,
            42,
            workspace="workspace",
            working_directory=self.workdir,
        )

    def test_available(self) -> None:
        """Test that available() returns True if incus is available."""
        self.mock_is_command_available(
            {
                "incus": True,
                "autopkgtest-virt-incus": True,
            }
        )
        self.run_incus_cmd.return_value = "environment:\n  driver: lxc | qemu\n"
        self.assertTrue(self.executor_cls.available())
        self.run_incus_cmd.assert_called_once_with("info")

    def test_available_no_incus(self) -> None:
        """Test that available() returns False if incus is not available."""
        self.mock_is_command_available(
            {
                "incus": False,
                "autopkgtest-virt-incus": True,
            }
        )
        self.run_incus_cmd.return_value = "environment:\n  driver: lxc | qemu\n"
        self.assertFalse(self.executor_cls.available())

    def test_available_no_autopkgtest_virt_incus(self) -> None:
        """Not available() if autopkgtest-virt-incus is not available."""
        self.mock_is_command_available(
            {
                "incus": True,
                "autopkgtest-virt-incus": False,
            }
        )
        self.run_incus_cmd.return_value = "environment:\n  driver: lxc | qemu\n"
        self.assertFalse(self.executor_cls.available())

    def test_available_no_incus_permission(self) -> None:
        """Test that available() returns False if the worker can't run incus."""
        self.mock_is_command_available(
            {
                "incus": True,
                "autopkgtest-virt-incus": True,
            }
        )
        self.run_incus_cmd.side_effect = CalledProcessError(
            cmd=("incus", "info"), returncode=1
        )
        self.assertFalse(self.executor_cls.available())

    def test_instantiation_fetches_artifact(self) -> None:
        """Test that instantiating the Executor fetches the artifact."""
        self.assertEqual(self.executor.system_image, self.image_artifact)

    def test_download_image(self) -> None:
        """Test that download_image calls the ImageCache.download_image."""
        self.image_artifact = self.mock_image_download(
            self.debusine_api, create_image=True, system_image=self.uses_images
        )

        def fake_run_incus_cmd(*args: str, **kwargs: Any) -> str:
            """Fake for run_incus_cmd."""
            if args == ("image", "info", "artifact/42"):
                raise CalledProcessError(cmd=("incus", *args), returncode=1)
            if args[:2] == ("image", "import"):
                return "Fingerprint: abc123\n"
            raise AssertionError(f"Unexpected args: {args}")

        self.run_incus_cmd.side_effect = fake_run_incus_cmd

        response = self.executor.download_image()

        self.run_incus_cmd.assert_any_call(
            "image", "info", "artifact/42", suppress_stderr=True
        )
        self.run_incus_cmd.assert_called_with(
            "image", "import", ANY, "--alias", "artifact/42"
        )
        expected_name = "artifact/42"
        self.assertEqual(response, expected_name)

    def test_image_name(self) -> None:
        """Test that image_name returns the expected image name."""
        response = self.executor.image_name()
        expected_name = "artifact/42"
        self.assertEqual(response, expected_name)

    def test_autopkgtest_virt_server(self) -> None:
        """Test that autopkgtest_virt_server returns incus."""
        self.assertEqual(self.executor.autopkgtest_virt_server(), "incus")

    def test_create(self) -> None:
        """Test create() return IncusInstance instance."""
        incus_instance = self.executor.create()

        self.assertIsInstance(incus_instance, IncusInstance)

    @patch("debusine.tasks.executors.incus.run_incus_cmd")
    def test_clean_up_image(self, run_incus_cmd: MagicMock) -> None:
        """Test that IncusImageCache.clean_up_image removes incus images."""
        assert issubclass(self.executor_cls, IncusExecutorMixin)
        self.executor_cls.clean_up_image(42)
        run_incus_cmd.assert_called_once_with("image", "rm", "artifact/42")

    def test_get_statistics_without_instance(self) -> None:
        """`get_statistics` returns empty statistics without an instance."""
        self.assertEqual(
            self.executor_cls.get_statistics(None), ExecutorStatistics()
        )

    def test_get_statistics_with_instance(self) -> None:
        """`get_statistics` returns all statistics with an instance."""
        instance = IncusInstance("image-name")
        instance._instance = "instance"
        run_result = MagicMock()
        run_result.returncode = 0
        run_result.stdout = dedent(
            """\
            # HELP incus_cpu_effective_total The total number of effective CPUs.
            # TYPE incus_cpu_effective_total gauge
            incus_cpu_effective_total{name="instance"} 2
            incus_cpu_effective_total{name="other-instance"} 4
            # HELP incus_filesystem_avail_bytes The number of available space in bytes.
            # TYPE incus_filesystem_avail_bytes gauge
            incus_filesystem_avail_bytes{mountpoint="/",name="instance"} 1e+10
            incus_filesystem_avail_bytes{mountpoint="/build",name="instance"} 6.7e+10
            incus_filesystem_avail_bytes{mountpoint="/",name="other-instance"} 1e+10
            # HELP incus_filesystem_size_bytes The size of the filesystem in bytes.
            # TYPE incus_filesystem_size_bytes gauge
            incus_filesystem_size_bytes{mountpoint="/",name="instance"} 2e+10
            incus_filesystem_size_bytes{mountpoint="/build",name="instance"} 7.4e+10
            incus_filesystem_size_bytes{mountpoint="/",name="other-instance"} 2e+10
            # HELP incus_memory_MemFree_bytes The amount of free memory.
            # TYPE incus_memory_MemFree_bytes gauge
            incus_memory_MemFree_bytes{name="instance"} 2.0e+9
            incus_memory_MemFree_bytes{name="other-instance"} 3.1e+9
            # HELP incus_memory_MemTotal_bytes The amount of used memory.
            # TYPE incus_memory_MemTotal_bytes gauge
            incus_memory_MemTotal_bytes{name="instance"} 2.1e+9
            incus_memory_MemTotal_bytes{name="other-instance"} 3.2e+9
            # HELP incus_warnings_total The number of active warnings.
            # TYPE incus_warnings_total counter
            incus_warnings_total 0
            """  # noqa: E501
        )

        with patch(
            "subprocess.run", autospec=True, return_value=run_result
        ) as mock_run:
            statistics = self.executor_cls.get_statistics(instance)

        mock_run.assert_called_once_with(
            ["incus", "query", "/1.0/metrics"],
            capture_output=True,
            check=True,
            text=True,
        )
        self.assertEqual(
            statistics,
            ExecutorStatistics(
                memory=100_000_000,
                disk_space=17_000_000_000,
                available_memory=2_100_000_000,
                available_disk_space=94_000_000_000,
                cpu_count=2,
            ),
        )

    def test_get_statistics_with_instance_limited_details(self) -> None:
        """`get_statistics` only returns statistics that are available."""
        instance = IncusInstance("image-name")
        instance._instance = "instance"
        run_result = MagicMock()
        run_result.returncode = 0
        run_result.stdout = dedent(
            """\
            # HELP incus_cpu_effective_total The total number of effective CPUs.
            # TYPE incus_cpu_effective_total gauge
            incus_cpu_effective_total{name="instance"} 2
            """  # noqa: E501
        )

        with patch(
            "subprocess.run", autospec=True, return_value=run_result
        ) as mock_run:
            statistics = self.executor_cls.get_statistics(instance)

        mock_run.assert_called_once_with(
            ["incus", "query", "/1.0/metrics"],
            capture_output=True,
            check=True,
            text=True,
        )
        self.assertEqual(statistics, ExecutorStatistics(cpu_count=2))


class IncusLXCExecutorTests(IncusExecutorCommonTests):
    """Unit tests for IncusLXCExecutor."""

    executor_cls = IncusLXCExecutor
    uses_images = False

    def test_backend_name(self) -> None:
        """Test that the backend_name attribute was set."""
        self.assertEqual(self.executor_cls.backend_name, BackendType.INCUS_LXC)

    def test_executor_class_finds_incus_lxc(self) -> None:
        """Test that executor_class() supports incus-lxc."""
        instance = executor_class(BackendType.INCUS_LXC)
        self.assertEqual(instance, self.executor_cls)

    def test_available_no_driver(self) -> None:
        """Test available() returns False if the lxc driver isn't there."""
        self.mock_is_command_available(
            {
                "incus": True,
                "autopkgtest-virt-incus": True,
            }
        )
        self.run_incus_cmd.return_value = "environment:\n  driver: qemu\n"
        self.assertFalse(self.executor_cls.available())

    def test_available_has_driver(self) -> None:
        """Test available() returns True if the lxc driver is there."""
        self.mock_is_command_available(
            {
                "incus": True,
                "autopkgtest-virt-incus": True,
            }
        )
        self.run_incus_cmd.return_value = "environment:\n  driver: lxc\n"
        self.assertTrue(self.executor_cls.available())

    def test_autopkgtest_virt_args(self) -> None:
        """Test that autopkgtest_virt_args returns sane arguments."""
        self.assertEqual(
            self.executor.autopkgtest_virt_args(),
            ["artifact/42", "--", "--profile", "debusine"],
        )


class IncusVMExecutorTests(IncusExecutorCommonTests):
    """Unit tests for IncusVMExecutor."""

    executor_cls = IncusVMExecutor
    uses_images = True

    def test_backend_name(self) -> None:
        """Test that the backend_name attribute was set."""
        self.assertEqual(self.executor_cls.backend_name, BackendType.INCUS_VM)

    def test_executor_class_finds_incus_vm(self) -> None:
        """Test that executor_class() supports incus-vm."""
        instance = executor_class(BackendType.INCUS_VM)
        self.assertEqual(instance, self.executor_cls)

    def test_available_no_driver(self) -> None:
        """Test available() returns False if the qemu driver isn't there."""
        self.mock_is_command_available({"incus": True})
        self.run_incus_cmd.return_value = "environment:\n  driver: lxc\n"
        self.assertFalse(self.executor_cls.available())

    def test_available_has_driver(self) -> None:
        """Test available() returns True if the qemu driver is there."""
        self.mock_is_command_available(
            {
                "incus": True,
                "autopkgtest-virt-incus": True,
            }
        )
        self.run_incus_cmd.return_value = "environment:\n  driver: qemu\n"
        self.assertTrue(self.executor_cls.available())

    def test_autopkgtest_virt_args(self) -> None:
        """Test that autopkgtest_virt_args returns sane arguments."""
        self.assertEqual(
            self.executor.autopkgtest_virt_args(),
            ["--vm", "artifact/42", "--", "--profile", "debusine"],
        )


class IncusInstanceTests(TestCase):
    """Tests for IncusInstance class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.instance = IncusInstance("image-name")

        run_incus_cmd_patcher = patch(
            "debusine.tasks.executors.incus.run_incus_cmd"
        )
        self.run_incus_cmd = run_incus_cmd_patcher.start()
        self.addCleanup(run_incus_cmd_patcher.stop)

        def fake_run_incus_cmd(*args: str) -> str:
            """
            Fake for run_incus_cmd.

            Pretend to operate normally.
            """
            if args[0] == "info":  # _generate_instance_name() probes
                raise CalledProcessError(cmd=("incus", *args), returncode=1)
            elif args[0] == "launch":
                return "Launching\n"
            elif args[0] == "stop":
                return ""
            elif args[0] == "restart":
                return ""
            elif args[0] == "file":  # push/pull
                return ""
            raise AssertionError(f"Unexpected args: {args}")

        self.run_incus_cmd.side_effect = fake_run_incus_cmd

    def patch_spin_until_started(self) -> None:
        """Patch _spin_until_started() out for testing."""
        patcher = patch.object(self.instance, "_spin_until_started")
        self._spin_until_started_mock = patcher.start()
        self.addCleanup(patcher.stop)

    def test_generate_instance_name(self) -> None:
        """Test that _generate_instance_name() picks an unused name."""
        rejected = None

        def fake_run_incus_cmd(*args: str) -> str:
            """
            Fake for run_incus_cmd.

            Reject the first name tested, accept the second.
            """
            nonlocal rejected
            if args[0] != "info":
                raise AssertionError(f"Unexpected args: {args}")
            if rejected:
                raise CalledProcessError(cmd=("incus", *args), returncode=1)
            else:
                rejected = args[1]
                return "Fingerprint: abc123\n"

        self.run_incus_cmd.side_effect = fake_run_incus_cmd

        instance_name = IncusInstance._generate_instance_name()

        self.assertIsNotNone(rejected)
        self.assertNotEqual(rejected, instance_name)
        calls = self.run_incus_cmd.call_args_list
        self.assertEqual(len(calls), 2)
        self.assertEqual(calls[0], call("info", rejected))
        self.assertEqual(calls[1], call("info", instance_name))

    def test_start(self) -> None:
        """Test that start() launches an instance."""
        self.patch_spin_until_started()
        self.instance.start()

        # First call would have been generating an instance name
        self.assertEqual(len(self.run_incus_cmd.call_args_list), 2)
        self.run_incus_cmd.assert_called_with(
            "launch", "--ephemeral", "--profile", "debusine", "image-name", ANY
        )
        self.assertIsNotNone(self.instance._instance)
        self._spin_until_started_mock.assert_called_once_with()

    @patch("subprocess.run")
    def test_spin_until_started(self, run: MagicMock) -> None:
        """Test that spin_until_started() spins until it succeeds."""
        count = 0

        def fake_run(args: list[str], **kwargs: Any) -> CompletedProcess[str]:
            """
            Fake for subprocess.run().

            Pretend to spin once.
            """
            nonlocal count
            returncode = 0
            stdout = ""
            stderr = ""
            if count == 0:
                returncode = 1
                stderr = "Error: VM agent isn't currently running\n"
            elif count == 1:
                returncode = 1
                stderr = (
                    "Failed to connect to system scope bus via local "
                    "transport: No such file or directory\n"
                )
            elif count == 2:
                returncode = 1
                stdout = "initializing\n"
            elif count == 3:
                returncode = 1
                stdout = "starting\n"
            else:
                stdout = "running\n"
            count += 1
            return CompletedProcess(
                args=args,
                returncode=returncode,
                stdout=stdout,
                stderr=stderr,
            )

        run.side_effect = fake_run

        self.instance._instance = "fake-instance"
        self.instance._spin_until_started(sleep_time=0, timeout=2)

        run.assert_called_with(
            [
                "incus",
                "exec",
                self.instance._instance,
                "systemctl",
                "is-system-running",
            ],
            capture_output=True,
            text=True,
        )

    @patch("subprocess.run")
    def test_spin_until_started_unexpected_failure(
        self, run: MagicMock
    ) -> None:
        """Test that spin_until_started() may raise InstanceNotRunning."""
        count = 0

        def fake_run(args: list[str], **kwargs: Any) -> CompletedProcess[str]:
            """
            Fake for subprocess.run().

            Pretend to spin once.
            """
            nonlocal count
            returncode = 0
            stdout = ""
            stderr = ""
            if count == 0:
                returncode = 1
                stderr = "Something weird"
            else:
                returncode = 0
                stdout = "running\n"
            count += 1
            return CompletedProcess(
                args=args,
                returncode=returncode,
                stdout=stdout,
                stderr=stderr,
            )

        run.side_effect = fake_run

        self.instance._instance = "fake-instance"

        with self.assertLogsContains(
            "Unexpected response from exec systemctl is-system-running: "
            "returncode: 1, stdout: , stderr: Something weird",
            logger=log,
        ):
            self.instance._spin_until_started(sleep_time=0, timeout=5)

    def test_spin_until_started_times_out(self) -> None:
        """Test that spin_until_started() times out."""
        self.instance._instance = "fake-instance"
        with self.assertRaisesRegex(
            InstanceNotRunning,
            "Incus Instance fake-instance hasn't booted or is missing "
            "incus-agent",
        ):
            self.instance._spin_until_started(timeout=0)

    def test_restart(self) -> None:
        """Test that restart() restarts an instance."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance
        self.run_incus_cmd.reset_mock()

        self.instance.restart()

        self.run_incus_cmd.assert_called_with("restart", instance_name)
        self.assertIsNotNone(self.instance._instance)

    def test_stop(self) -> None:
        """Test that stop() stops an instance."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance
        self.run_incus_cmd.reset_mock()

        self.instance.stop()

        self.run_incus_cmd.assert_called_with("stop", instance_name)
        self.assertIsNone(self.instance._instance)

    def test_file_push(self) -> None:
        """Test that file_push() copies a file into an instance."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance
        self.run_incus_cmd.reset_mock()

        self.instance.file_push(Path("/some/input"), Path("/tmp/destination"))

        self.run_incus_cmd.assert_called_with(
            "file",
            "push",
            "--uid",
            "0",
            "--gid",
            "0",
            "--mode",
            "644",
            "/some/input",
            f"{instance_name}/tmp/destination",
        )

    def test_directory_push(self) -> None:
        """Test that directory_push() copies a directory into an instance."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance
        self.run_incus_cmd.reset_mock()
        self.run_incus_cmd.side_effect = None

        self.instance.directory_push(Path("/some/directory"), Path("/tmp"))

        self.run_incus_cmd.assert_has_calls(
            [
                call(
                    "file",
                    "push",
                    "--recursive",
                    "/some/directory",
                    f"{instance_name}/tmp",
                ),
                call(
                    "exec",
                    instance_name,
                    "--",
                    "chown",
                    "-R",
                    "0:0",
                    "/tmp/directory",
                ),
            ]
        )

    def test_file_pull(self) -> None:
        """Test that file_pull() copies a file into an instance."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance
        self.run_incus_cmd.reset_mock()

        self.instance.file_pull(Path("/etc/hosts"), Path("/tmp/destination"))

        self.run_incus_cmd.assert_called_with(
            "file", "pull", f"{instance_name}/etc/hosts", "/tmp/destination"
        )

    @patch("subprocess.run")
    def test_run(self, run: MagicMock) -> None:
        """Test that run() executes commands as non-root."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance
        self.run_incus_cmd.reset_mock()

        self.instance._uid_map["_debusine"] = 1042

        self.instance.run(["true"])

        run.assert_called_with(
            ["incus", "exec", instance_name, "--user", "1042", "--", "true"],
            text=None,
            stdout=PIPE,
            stderr=PIPE,
        )

    @patch("subprocess.run")
    def test_run_cwd_absolute(self, run: MagicMock) -> None:
        """run() can execute commands in an absolute working directory."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance
        self.run_incus_cmd.reset_mock()

        self.instance._uid_map["_debusine"] = 1042

        self.instance.run(["true"], cwd="/home/build/work")

        run.assert_called_with(
            [
                "incus",
                "exec",
                instance_name,
                "--cwd",
                "/home/build/work",
                "--user",
                "1042",
                "--",
                "true",
            ],
            text=None,
            stdout=PIPE,
            stderr=PIPE,
        )

    @patch("subprocess.run")
    def test_run_cwd_relative(self, run: MagicMock) -> None:
        """run() can execute commands in a relative working directory."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance
        self.run_incus_cmd.reset_mock()

        self.instance._uid_map["_debusine"] = 1042

        self.instance.run(["true"], cwd="work")

        run.assert_called_with(
            [
                "incus",
                "exec",
                instance_name,
                "--cwd",
                "/root/work",
                "--user",
                "1042",
                "--",
                "true",
            ],
            text=None,
            stdout=PIPE,
            stderr=PIPE,
        )

    @patch("subprocess.run")
    def test_run_as_root(self, run: MagicMock) -> None:
        """Test that run() executes commands as root."""
        self.patch_spin_until_started()
        self.instance.start()
        instance_name = self.instance._instance

        self.instance.run(["true", "--help"], run_as_root=True)

        run.assert_called_with(
            ["incus", "exec", instance_name, "--", "true", "--help"],
            text=None,
            stdout=PIPE,
            stderr=PIPE,
        )


# Avoid running tests from common base classes.
del IncusImageCacheCommonTests
del IncusExecutorCommonTests



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Test of the signon authentication middleware for external providers."""

from typing import override

from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured, MiddlewareNotUsed
from django.http import HttpRequest, HttpResponse
from django.test import RequestFactory, TestCase, override_settings

from debusine.server.signon import providers
from debusine.server.signon.middleware import (
    RequestSignonProtocol,
    SignonMiddleware,
)
from debusine.server.signon.signon import Signon

PROVIDERS_CONFIG = [
    providers.Provider(
        name="debsso",
        label="sso.debian.org",
    ),
    providers.Provider(name="salsa", label="Salsa"),
]


def mock_get_response(
    request: HttpRequest,  # noqa: ARG001, U100
) -> HttpResponse:
    """Mock version of get_response to use for testing middlewares."""
    return HttpResponse()


class SignonSubclass(Signon):
    """Subclass of Signon to use in tests."""


class TestMiddleware(TestCase):
    """Test SignonMiddleware."""

    @override
    def setUp(self) -> None:
        """Provide a mock unauthenticated request for tests."""
        super().setUp()
        self.factory = RequestFactory()
        self.request = self.factory.get("/")

    @override_settings(SIGNON_PROVIDERS=PROVIDERS_CONFIG)
    def test_no_authentication_middleware(self) -> None:
        """Enforce that AuthenticationMiddleware is required."""
        mw = SignonMiddleware(get_response=mock_get_response)
        with self.assertRaises(ImproperlyConfigured):
            mw(self.request)

    @override_settings(SIGNON_PROVIDERS=[])
    def test_no_providers(self) -> None:
        """Skip middleware if there are no providers."""
        with self.assertRaises(MiddlewareNotUsed):
            SignonMiddleware(get_response=mock_get_response)

    @override_settings(SIGNON_PROVIDERS=PROVIDERS_CONFIG)
    def test_all_fine(self) -> None:
        """Test the case where all requirements are met."""
        self.request.user = AnonymousUser()
        assert not isinstance(self.request, RequestSignonProtocol)

        mw = SignonMiddleware(get_response=mock_get_response)
        mw(self.request)

        assert isinstance(self.request, RequestSignonProtocol)
        self.assertIsInstance(self.request.signon, Signon)

    @override_settings(
        SIGNON_PROVIDERS=PROVIDERS_CONFIG,
        SIGNON_CLASS="debusine.does.not.Exist",
    )
    def test_signon_class_not_found(self) -> None:
        """Test setting SIGNON_CLASS to an invalid path."""
        with self.assertRaisesRegex(
            ImportError, r"No module named 'debusine.does'"
        ):
            SignonMiddleware(get_response=mock_get_response)

    @override_settings(
        SIGNON_PROVIDERS=PROVIDERS_CONFIG,
        SIGNON_CLASS="debusine.db.models.Scope",
    )
    def test_signon_class_invalid(self) -> None:
        """Test setting SIGNON_CLASS to a non-Signon class."""
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            r"debusine.db.models.Scope is not a subclass of Signon",
        ):
            SignonMiddleware(get_response=mock_get_response)

    @override_settings(
        SIGNON_PROVIDERS=PROVIDERS_CONFIG,
        SIGNON_CLASS=f"{SignonSubclass.__module__}.SignonSubclass",
    )
    def test_signon_class(self) -> None:
        """Test setting SIGNON_CLASS to a valid subclass."""
        mw = SignonMiddleware(get_response=mock_get_response)
        self.assertIs(mw.signon_class, SignonSubclass)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""DebusineBaseCommand extends BaseCommand with extra functionality."""

import signal
import types
from typing import Any, NoReturn, override

import yaml
from django.core.management import BaseCommand, CommandError
from django.db.utils import DatabaseError
from yaml import YAMLError


class DebusineBaseCommand(BaseCommand):
    """Extend Django Base Command with functionality used by Debusine."""

    @staticmethod
    def _exit_handler(signum: int, frame: types.FrameType | None) -> NoReturn:
        """
        Exit without printing Python's default stack trace.

        A user can Control+C and debusine-admin does not print all the
        stack trace. This could happen in any command but is more obvious
        on the interactive commands (such as `remove_tokens` when it asks for
        confirmation)
        """
        signum, frame  # fake usage for vulture
        raise SystemExit(3)

    def cleanup_arguments(self, *args: Any, **options: Any) -> None:
        """
        Clean up objects created by parsing arguments.

        For example, this can be used to close files created by
        :py:class:`argparse.FileType`.
        """

    @override
    def execute(self, *args: Any, **options: Any) -> None:
        """
        Debusine BaseCommand common functionality.

        - Catch OperationError exceptions to print a helpful message
        - Set self.verbosity for --verbosity/-v option

        Possible OperationErrors: database not reachable, invalid database
        credentials, etc.
        """
        try:
            signal.signal(signal.SIGINT, self._exit_handler)
            signal.signal(signal.SIGTERM, self._exit_handler)

            self.verbosity = options["verbosity"]

            super().execute(*args, **options)
        except DatabaseError as exc:
            raise CommandError(f"Database error: {exc}", returncode=3)
        finally:
            self.cleanup_arguments(*args, **options)

    def print_verbose(self, msg: str) -> None:
        r"""Write msg + "\n" to self.stdout if self.verbosity > 1."""
        if self.verbosity > 1:
            self.stdout.write(msg)
            self.stdout.flush()

    def parse_yaml_data(self, data_yaml: str) -> Any:
        """Parse data_yaml. If not valid raise CommandError."""
        try:
            data = yaml.safe_load(data_yaml)
        except YAMLError as err:
            raise CommandError(f"Error parsing YAML: {err}", returncode=3)

        return data



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command worker_pool."""

from io import StringIO
from typing import Any
from unittest import mock

import yaml
from django.core.management.base import CommandError
from django.test import override_settings

from debusine.assets import CloudProvidersType, DummyProviderAccountData
from debusine.db.models import Worker, WorkerPool
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.server.management.commands.worker_pool import Command
from debusine.server.management.management_utils import sort_keys
from debusine.server.worker_pools import (
    DummyWorkerPool,
    DummyWorkerPoolSpecification,
    WorkerPoolLimits,
)
from debusine.test.django import TestCase
from debusine.utils.input import YamlEditor


@override_settings(LANGUAGE_CODE="en-us")
class WorkerPoolCommandTests(TabularOutputTests, TestCase):
    """Tests for the `worker_pool` command."""

    def test_create_defaults_from_stdin(self) -> None:
        name = "test"
        provider_account = self.playground.create_cloud_provider_account_asset()
        provider_account_data = provider_account.data_model
        assert isinstance(provider_account_data, DummyProviderAccountData)
        specifications = DummyWorkerPoolSpecification(features=["foo"])

        stdout, stderr, exit_code = call_command(
            "worker_pool",
            "create",
            name,
            "--provider-account",
            provider_account_data.name,
            "--architectures",
            "amd64,i386",
            "--specifications",
            "-",
            stdin=StringIO(yaml.safe_dump(specifications.model_dump())),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        pool = WorkerPool.objects.get(name=name)
        self.assertTrue(pool.enabled)
        self.assertEqual(pool.architectures, ["amd64", "i386"])
        self.assertEqual(pool.tags, [])
        self.assertEqual(pool.provider_account, provider_account)
        self.assertEqual(pool.specifications, specifications.model_dump())
        self.assertTrue(pool.instance_wide)
        self.assertFalse(pool.ephemeral)
        self.assertEqual(pool.limits, WorkerPoolLimits().model_dump())

    def test_create_all_args_from_files(self) -> None:
        name = "test"
        provider_account = self.playground.create_cloud_provider_account_asset()
        provider_account_data = provider_account.data_model
        assert isinstance(provider_account_data, DummyProviderAccountData)
        specifications = DummyWorkerPoolSpecification(features=["something"])
        specifications_file = self.create_temporary_file(
            contents=specifications.model_dump_json().encode()
        )

        limits = WorkerPoolLimits(max_idle_seconds=60)
        limits_file = self.create_temporary_file(
            contents=limits.model_dump_json().encode()
        )
        apt_authentication = self.playground.create_apt_authentication_asset(
            workspace=None
        )

        stdout, stderr, exit_code = call_command(
            "worker_pool",
            "create",
            name,
            "--provider-account",
            provider_account_data.name,
            "--no-enabled",
            "--architectures",
            "i386",
            "--tags",
            "tag-1",
            "--specifications",
            str(specifications_file),
            "--no-instance-wide",
            "--ephemeral",
            "--limits",
            str(limits_file),
            "--apt-authentication",
            apt_authentication.data["name"],
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        pool = WorkerPool.objects.get(name=name)
        self.assertFalse(pool.enabled)
        self.assertEqual(pool.architectures, ["i386"])
        self.assertEqual(pool.tags, ["tag-1"])
        self.assertEqual(pool.provider_account, provider_account)
        self.assertEqual(pool.specifications, specifications.model_dump())
        self.assertFalse(pool.instance_wide)
        self.assertTrue(pool.ephemeral)
        self.assertEqual(pool.limits, limits.model_dump())
        self.assertEqual(pool.apt_authentication, apt_authentication)

    def test_create_idempotent(self) -> None:
        pool = self.playground.create_worker_pool()
        specifications_file = self.create_temporary_file(
            contents=yaml.safe_dump(pool.specifications).encode()
        )
        limits = WorkerPoolLimits(max_idle_seconds=60)
        limits_file = self.create_temporary_file(
            contents=limits.model_dump_json().encode()
        )
        stdout, stderr, exit_code = call_command(
            "worker_pool",
            "create",
            pool.name,
            "--provider-account",
            pool.provider_account.data["name"],
            "--no-enabled",
            "--architectures",
            "i386",
            "--tags",
            "tag-1",
            "--specifications",
            str(specifications_file),
            "--no-instance-wide",
            "--ephemeral",
            "--limits",
            str(limits_file),
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        pool.refresh_from_db()

        self.assertFalse(pool.enabled)
        self.assertEqual(pool.architectures, ["i386"])
        self.assertEqual(pool.tags, ["tag-1"])
        self.assertFalse(pool.instance_wide)
        self.assertTrue(pool.ephemeral)
        self.assertEqual(pool.limits, limits.model_dump())

    def test_create_invalid_specifications(self) -> None:
        name = "test"
        provider_account = self.playground.create_cloud_provider_account_asset()
        provider_account_data = provider_account.data_model
        assert isinstance(provider_account_data, DummyProviderAccountData)
        specifications_file = self.create_temporary_file(
            contents=yaml.dump(
                {
                    "provider_type": "dummy",
                    "something": "some value",
                },
                encoding="utf-8",
            )
        )
        with self.assertRaisesRegex(
            CommandError, r"(?s)Error creating worker_pool:.*validation error"
        ):
            call_command(
                "worker_pool",
                "create",
                name,
                "--provider-account",
                provider_account_data.name,
                "--specifications",
                str(specifications_file),
                "--architectures",
                "amd64",
            )

    def test_create_invalid_limits(self) -> None:
        name = "test"
        provider_account = self.playground.create_cloud_provider_account_asset()
        provider_account_data = provider_account.data_model
        assert isinstance(provider_account_data, DummyProviderAccountData)
        specifications = DummyWorkerPoolSpecification(features=["something"])
        specifications_file = self.create_temporary_file(
            contents=specifications.model_dump_json().encode()
        )
        limits_file = self.create_temporary_file(
            contents=yaml.dump(
                {
                    "something": "some value",
                },
                encoding="utf-8",
            )
        )
        with self.assertRaisesRegex(
            CommandError, r"(?s)Error creating worker_pool:.*validation error"
        ):
            call_command(
                "worker_pool",
                "create",
                name,
                "--provider-account",
                provider_account_data.name,
                "--specifications",
                str(specifications_file),
                "--limits",
                str(limits_file),
                "--architectures",
                "amd64",
            )

    def test_create_unknown_provider_account(self) -> None:
        with self.assertRaisesRegex(
            CommandError, r"Cloud provider account asset 'unknown' not found"
        ):
            call_command(
                "worker_pool",
                "create",
                "test",
                "--provider-account",
                "unknown",
            )

    def test_create_unknown_apt_authentication(self) -> None:
        with self.assertRaisesRegex(
            CommandError,
            r"APT authentication asset with slug ':unknown' not found",
        ):
            call_command(
                "worker_pool",
                "create",
                "test",
                "--apt-authentication",
                "unknown",
            )

    def test_delete(self) -> None:
        pool = self.playground.create_worker_pool()
        stdout, stderr, exit_code = call_command(
            "worker_pool",
            "delete",
            pool.name,
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertFalse(WorkerPool.objects.filter(name="test").exists())

    def test_delete_not_found(self) -> None:
        with self.assertRaisesRegex(
            CommandError, r"Worker Pool 'unknown' not found"
        ):
            call_command(
                "worker_pool",
                "delete",
                "unknown",
            )

    def test_delete_running_workers(self) -> None:
        pool = self.playground.create_worker_pool()
        self.playground.create_worker(worker_pool=pool)
        with self.assertRaisesRegex(
            CommandError, r"Worker pool 'test' has 1 running workers\."
        ):
            call_command(
                "worker_pool",
                "delete",
                "test",
            )

    def test_delete_running_workers_force(self) -> None:
        pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=pool)
        assert worker.activation_token is not None
        self.addCleanup(worker.activation_token.delete)
        with mock.patch.object(
            DummyWorkerPool, "terminate_worker"
        ) as mock_terminate_worker:
            stdout, stderr, exit_code = call_command(
                "worker_pool",
                "delete",
                "test",
                "--force",
            )
        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)
        mock_terminate_worker.assert_called_once_with(worker)
        self.assertFalse(WorkerPool.objects.filter(name="test").exists())
        self.assertFalse(Worker.objects.filter(id=worker.id).exists())

    def test_enable(self) -> None:
        pool = self.playground.create_worker_pool(enabled=False)
        stdout, stderr, exit_code = call_command(
            "worker_pool",
            "enable",
            pool.name,
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        pool.refresh_from_db()
        self.assertTrue(pool.enabled)

    def test_disable(self) -> None:
        pool = self.playground.create_worker_pool()
        stdout, stderr, exit_code = call_command(
            "worker_pool",
            "disable",
            pool.name,
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        pool.refresh_from_db()
        self.assertFalse(pool.enabled)

    def test_list(self) -> None:
        dummy_provider = self.playground.create_cloud_provider_account_asset(
            name="dummy"
        )
        aws_provider = self.playground.create_cloud_provider_account_asset(
            name="aws", cloud_provider=CloudProvidersType.AWS
        )
        pools = [
            self.playground.create_worker_pool(
                name="dummy", provider_account=dummy_provider
            ),
            self.playground.create_worker_pool(
                name="aws", provider_account=aws_provider
            ),
        ]
        expected_data = [
            {
                "name": pool.name,
                "provider_account": pool.provider_account.data["name"],
                "enabled": pool.enabled,
                "specifications": sort_keys(
                    pool.specifications_model.model_dump(
                        mode="json", exclude_unset=True
                    )
                ),
                "limits": sort_keys(
                    pool.limits_model.model_dump(
                        mode="json", exclude_unset=True
                    )
                ),
            }
            for pool in pools
        ]

        with self.assertPrintsTable() as output:
            stdout, stderr, exit_code = call_command("worker_pool", "list")
        self.assertEqual(exit_code, 0)
        self.assertEqual(
            output.col(0), [expected["name"] for expected in expected_data]
        )
        self.assertEqual(
            output.col(1),
            [expected["provider_account"] for expected in expected_data],
        )
        self.assertEqual(
            output.col(2),
            [str(expected["enabled"]) for expected in expected_data],
        )
        self.assertEqual(
            output.col(3),
            [str(expected["specifications"]) for expected in expected_data],
        )
        self.assertEqual(
            output.col(4),
            [str(expected["limits"]) for expected in expected_data],
        )

        stdout, stderr, exit_code = call_command(
            "worker_pool", "list", "--yaml"
        )
        self.assertEqual(exit_code, 0)
        self.assertEqual(stderr, "")
        data = yaml.safe_load(stdout)
        self.assertEqual(data, expected_data)

    def test_launch(self) -> None:
        pool = self.playground.create_worker_pool()
        stdout, stderr, exit_code = call_command(
            "worker_pool",
            "launch",
            pool.name,
            "2",
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertEqual(
            pool.workers_running.filter(
                worker_pool_data__launched=True
            ).count(),
            2,
        )

    def test_terminate(self) -> None:
        pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=pool)
        worker.worker_pool_data = {"launched": True}
        worker.save()
        self.assertIsNotNone(worker.instance_created_at)

        stdout, stderr, exit_code = call_command(
            "worker_pool",
            "terminate",
            worker.name,
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        worker.refresh_from_db()
        self.assertIsNone(worker.instance_created_at)
        self.assertIsNone(worker.worker_pool_data)

    def test_terminate_non_pool_worker(self) -> None:
        worker = self.playground.create_worker()
        with self.assertRaisesRegex(
            CommandError, r"Worker 'computer-lan' is not a member of a pool\."
        ):
            call_command(
                "worker_pool",
                "terminate",
                worker.name,
            )

    def test_terminate_unknown_worker(self) -> None:
        with self.assertRaisesRegex(
            CommandError, r"Worker 'unknown' not found"
        ):
            call_command(
                "worker_pool",
                "terminate",
                "unknown",
            )

    def test_unexpected_action(self) -> None:
        command = Command()

        with self.assertRaisesRegex(
            CommandError, r"Action 'does_not_exist' not found"
        ) as exc:
            command.handle(action="does_not_exist")

        self.assertEqual(getattr(exc.exception, "returncode"), 3)

    def call_edit(
        self,
        *args: str,
        success: bool = True,
        output: dict[str, Any] | None = None,
    ) -> tuple[str, str, int | str]:
        """Call edit, mocking the editor with success and output data."""

        def mock_edit(self_: YamlEditor[dict[str, Any]]) -> bool:
            if output is not None:
                self_.value = output
            self_.cleanup()
            return success

        with mock.patch(
            "debusine.utils.input.YamlEditor.edit",
            side_effect=mock_edit,
            autospec=True,
        ):
            return call_command("worker_pool", "edit", *args)

    def test_edit_aborted(self) -> None:
        pool = self.playground.create_worker_pool()
        _, stderr, exit_code = self.call_edit(pool.name, success=False)
        self.assertEqual(exit_code, 1)
        self.assertEqual(stderr, "Edit aborted.\n")

    def test_edit_noop(self) -> None:
        pool = self.playground.create_worker_pool()
        _, stderr, exit_code = self.call_edit(pool.name)
        self.assertEqual(exit_code, 1)
        self.assertEqual(stderr, "No changes.\n")

    def test_edit_updates_only_changed_data(self) -> None:
        pool = self.playground.create_worker_pool()
        stdout, _, exit_code = self.call_edit(
            pool.name,
            output={
                "architectures": ["m68k"],
                "specifications": pool.specifications,
            },
        )
        self.assertEqual(exit_code, 0)
        self.assertEqual(stdout, "Updated architectures.\n")

        pool.refresh_from_db()
        self.assertEqual(pool.architectures, ["m68k"])

    def test_edit_updates_provider_account(self) -> None:
        pool = self.playground.create_worker_pool()
        provider_account = self.playground.create_cloud_provider_account_asset(
            name="different-provider"
        )
        stdout, _, exit_code = self.call_edit(
            pool.name,
            output={
                "provider_account": "different-provider",
            },
        )
        self.assertEqual(exit_code, 0)
        self.assertEqual(stdout, "Updated provider_account.\n")

        pool.refresh_from_db()
        self.assertEqual(pool.provider_account, provider_account)

    def test_edit_updates_apt_authentication(self) -> None:
        pool = self.playground.create_worker_pool()
        apt_authentication = self.playground.create_apt_authentication_asset(
            workspace=None
        )
        stdout, _, exit_code = self.call_edit(
            pool.name,
            output={
                "apt_authentication": apt_authentication.data["name"],
            },
        )
        self.assertEqual(exit_code, 0)
        self.assertEqual(stdout, "Updated apt_authentication.\n")

        pool.refresh_from_db()
        self.assertEqual(pool.apt_authentication, apt_authentication)

    def test_edit_removes_apt_authentication(self) -> None:
        pool = self.playground.create_worker_pool(
            apt_authentication=self.playground.create_apt_authentication_asset(
                workspace=None
            )
        )
        stdout, _, exit_code = self.call_edit(
            pool.name,
            output={
                "apt_authentication": None,
            },
        )
        self.assertEqual(exit_code, 0)
        self.assertEqual(stdout, "Updated apt_authentication.\n")

        pool.refresh_from_db()
        self.assertIsNone(pool.apt_authentication)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for MemoryFileBackend."""

from typing import override
from unittest import mock

from debusine.db.models import FileInStore, FileStore
from debusine.db.tests.utils import _calculate_hash_from_data
from debusine.server.file_backend.memory import (
    MemoryFileBackend,
    MemoryFileBackendEntry,
)
from debusine.test.django import TestCase
from debusine.test.test_utils import data_generator


class MemoryFileBackendTests(TestCase):
    """Tests for MemoryFileBackend."""

    @override
    def setUp(self) -> None:
        """Initialize file backend to be tested."""
        super().setUp()
        self.store = FileStore.objects.create(
            name="memory",
            backend=FileStore.BackendChoices.MEMORY,
            configuration={"name": "test_memory"},
        )
        self.backend = MemoryFileBackend(self.store)

    @override
    def tearDown(self) -> None:
        """Remove test memory storage."""
        MemoryFileBackend.storages.pop("test_memory", None)
        super().tearDown()

    def test_entry_str(self) -> None:
        """MemoryFileBackendEntry.__str__ returns the hex digest."""
        file = self.playground.create_file_in_backend(self.backend)
        entry = self.backend.get_entry(file)
        self.assertEqual(str(entry), file.sha256.hex())

    def test_get_local_path(self) -> None:
        """MemoryFileBackend.get_local_path returns None."""
        temp_file_path = self.create_temporary_file()
        file = self.backend.add_file(temp_file_path)
        local_path = self.backend.get_local_path(file)
        self.assertIsNone(local_path)

    def test_get_temporary_local_path(self) -> None:
        """MemoryFileBackend.get_temporary_local_path yields a local path."""
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = self.backend.add_file(temp_file_path)

        with self.backend.get_temporary_local_path(file) as temp_path:
            self.assertEqual(temp_path.read_bytes(), contents)

    def test_get_url(self) -> None:
        """MemoryFileBackend.get_url return None."""
        temp_file_path = self.create_temporary_file()
        file = self.backend.add_file(temp_file_path)
        url = self.backend.get_url(file)
        self.assertIsNone(url)

    def test_add_file(self) -> None:
        """MemoryFileBackend.add_file reuse the fileobj."""
        contents = b"This is a test"
        temp_file_path = self.create_temporary_file(contents=contents)

        file = self.backend.add_file(temp_file_path)

        self.assertEqual(file.size, len(contents))
        self.assertEqual(file.hash_digest, _calculate_hash_from_data(contents))
        self.assertTrue(
            FileInStore.objects.filter(
                store=self.backend.db_store, file=file
            ).exists()
        )
        self.assertEqual(
            self.backend.storages["test_memory"][file.hash_digest.hex()],
            contents,
        )

    def test_add_file_for_an_added_file(self) -> None:
        """add_file add a file that is already in the backend."""
        hexsum = (
            "c7be1ed902fb8dd4d48997c6452f5d7e509fbcdbe2808b16bcf4edce4c07d14e"
        )
        original_contents = b"This is a test"
        file_path = self.create_temporary_file(contents=original_contents)
        with mock.patch(
            "debusine.server.file_backend.memory.MemoryFileBackendEntry.add"
        ) as mock_add:
            self.backend.add_file(file_path)
        mock_add.assert_called_with(file_path)

        self.backend._store_file(hexsum, original_contents)
        with mock.patch(
            "debusine.server.file_backend.memory.MemoryFileBackendEntry.add"
        ) as mock_add:
            self.backend.add_file(file_path)
        mock_add.assert_not_called()

    def test_get_stream(self) -> None:
        """MemoryFileBackend.get_stream return a stream with file contents."""
        contents = b"this is a test"
        temp_file_path = self.create_temporary_file(contents=contents)
        file = self.backend.add_file(temp_file_path)

        with self.backend.get_stream(file) as f:
            self.assertEqual(f.read(), contents)

    def test_entry_remove(self) -> None:
        """MemoryFileBackendEntry.remove deletes the file from the backend."""
        temp_file_path = self.create_temporary_file()
        file = self.backend.add_file(temp_file_path)
        with self.backend.get_stream(file) as fd:
            fd.read()

        self.backend.get_entry(file).remove()

        with self.assertRaises(KeyError), self.backend.get_stream(file) as fd:
            pass  # pragma: no cover

    def test_entry_removes_twice(self) -> None:
        """
        MemoryFileBackendEntry.remove tries to delete a file that did not exist.

        This could happen if remove() was called concurrently.
        """
        temp_file_path = self.create_temporary_file()
        file = self.backend.add_file(temp_file_path)
        entry = self.backend.get_entry(file)
        entry.remove()
        entry.remove()

    def test_list_entries(self) -> None:
        """MemoryFileBackend.list_entries returns all entries in the backend."""
        data_gen = data_generator(16)
        files = [
            self.playground.create_file_in_backend(
                self.backend, contents=next(data_gen)
            )
            for _ in range(10)
        ]
        self.assertCountEqual(
            self.backend.list_entries(),
            [
                MemoryFileBackendEntry(
                    backend=self.backend, identifier=file.sha256.hex()
                )
                for file in files
            ],
        )



# Copyright 2016-2023 Enrico Zini <enrico@debian.org>
# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Helper functions for the signon module."""


def split_full_name(name: str) -> tuple[str, str]:
    """
    Arbitrary split a full name into (first_name, last_name).

    This is better than nothing, but not a lot better than that.
    """
    # See http://www.kalzumeus.com/2010/06/17/falsehoods-programmers-believe-about-names/  # noqa: E501
    fn = name.split()
    if len(fn) == 1:
        return fn[0], ""
    elif len(fn) == 2:
        return fn[0], fn[1]
    elif len(fn) == 3:
        return " ".join(fn[0:2]), fn[2]
    else:
        middle = len(fn) // 2
        return " ".join(fn[:middle]), " ".join(fn[middle:])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the make_signed_source workflow."""

from collections.abc import Sequence
from typing import ClassVar, override

from debusine.artifacts.models import (
    ArtifactCategory,
    DebusineSigningInput,
    DebusineSigningOutput,
    TaskTypes,
)
from debusine.assets import KeyPurpose
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, WorkRequest
from debusine.server.workflows import MakeSignedSourceWorkflow
from debusine.server.workflows.models import (
    BaseWorkflowData,
    MakeSignedSourceWorkflowData,
    MakeSignedSourceWorkflowDynamicData,
    SbuildWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import (
    SampleWorkflow,
    WorkflowTestCase,
)
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    BackendType,
    BaseDynamicTaskData,
    InputArtifactMultiple,
    LookupMultiple,
    SbuildInput,
)
from debusine.test.django import preserve_db_task_registry


class MakeSignedSourceWorkflowTests(WorkflowTestCase[MakeSignedSourceWorkflow]):
    """Unit tests for :py:class:`MakeSignedSourceWorkflow`."""

    SAMPLE_TASK_DATA = {
        "vendor": "debian",
        "codename": "trixie",
        "architectures": ["amd64"],
        "purpose": KeyPurpose("uefi"),
        "key": "ABC123",
    }

    source_artifact: ClassVar[Artifact]
    binary_amd64: ClassVar[Artifact]
    binary_i386: ClassVar[Artifact]
    upload_artifact: ClassVar[Artifact]
    signing_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_artifact = cls.playground.create_source_artifact(
            name="hello"
        )
        cls.binary_amd64 = (
            cls.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello", architecture="amd64"
            )
        )
        cls.binary_i386 = cls.playground.create_minimal_binary_package_artifact(
            srcpkg_name="hello", architecture="i386"
        )
        cls.upload_artifact = cls.playground.create_upload_artifacts(
            src_name="linux-base"
        ).upload
        cls.signing_artifact = cls.create_signing_template_artifact("amd64")

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["binary_artifacts"] = LookupMultiple(
            (self.binary_amd64.id, self.upload_artifact.id)
        )
        self.task_data["signing_template_artifacts"] = LookupMultiple(
            (self.signing_artifact.id,)
        )

    @classmethod
    def create_signing_template_artifact(cls, architecture: str) -> Artifact:
        """
        Create a signing template artifact.

        :param architecture: CollectionItem.data["architecture"]
        """
        artifact, _ = cls.playground.create_artifact(
            category=ArtifactCategory.BINARY_PACKAGE,
            data={
                "srcpkg_name": "hello",
                "srcpkg_version": "1.0.0",
                "deb_fields": {
                    "Package": "hello",
                    "Version": "1.0.0",
                    "Architecture": architecture,
                },
                "deb_control_files": [],
            },
        )
        return artifact

    def orchestrate(
        self,
        task_data: MakeSignedSourceWorkflowData,
        architectures: Sequence[str],
        create_sbuild_promises: bool,
        extra_sbuild_architectures: Sequence[str] | None = None,
    ) -> WorkRequest:
        """
        Create and orchestrate a MakeSignedSourceWorkflow.

        :param task_data: data for the workflow
        :param architectures: for each architecture, depending on "mode" param,
          creates an sbuild child promising a binary SIGNING_INPUT and
          BINARY_PACKAGE promising artifacts or only the artifacts (see
          create_promises / create_artifacts)
        :param create_sbuild_promises: if True, create an `sbuild`
          sub-workflow
        """
        source_artifact = self.playground.create_source_artifact()

        class ExamplePipeline(
            SampleWorkflow[BaseWorkflowData, BaseDynamicTaskData]
        ):
            """Pipeline workflow."""

            @override
            def populate(self) -> None:
                if create_sbuild_promises:
                    sbuild_architectures = [
                        *architectures,
                        *(extra_sbuild_architectures or []),
                    ]
                    sbuild, _ = self.work_request_ensure_child_workflow(
                        task_name="sbuild",
                        task_data=SbuildWorkflowData(
                            input=SbuildInput(
                                source_artifact=source_artifact.id
                            ),
                            target_distribution="debian:bookworm",
                            architectures=sbuild_architectures,
                            signing_template_names={
                                arch: [f"hello-{arch}-signed-template"]
                                for arch in sbuild_architectures
                            },
                        ),
                        workflow_data=WorkRequestWorkflowData(
                            display_name="sbuild", step="sbuild"
                        ),
                    )
                    self.orchestrate_child(sbuild)

                make_signed_source, _ = self.work_request_ensure_child_workflow(
                    task_name="make_signed_source",
                    task_data=task_data,
                    workflow_data=WorkRequestWorkflowData(
                        display_name="make signed source",
                        step="make_signed_source",
                    ),
                )
                self.orchestrate_child(make_signed_source)

        root = self.playground.create_workflow(task_name="examplepipeline")
        self.assertTrue(self.playground.orchestrate_workflow(root))

        return root

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            MakeSignedSourceWorkflowDynamicData(
                subject="hello linux-base",
                parameter_summary="hello linux-base for debian:trixie",
                binary_artifacts_ids=[
                    self.binary_amd64.id,
                    self.upload_artifact.id,
                ],
                signing_template_artifacts_ids=[self.signing_artifact.id],
            ),
        )

    def test_get_input_artifacts(self) -> None:
        workflow = self.task
        workflow.dynamic_data = workflow.compute_dynamic_data()
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[
                        self.binary_amd64.id,
                        self.upload_artifact.id,
                    ],
                ),
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.signing_template_artifacts,
                    label="signing_template_artifacts",
                    artifact_ids=[self.signing_artifact.id],
                ),
            ],
        )

    @preserve_db_task_registry()
    def test_populate(self) -> None:
        """Test populate."""
        # Expected architectures as per intersection of architectures
        # of signing_template_artifacts and binary_artifacts
        architectures = ["amd64", "i386"]

        # Create artifacts needed to satisfy task data lookups
        for arch in [*architectures, "arm64"]:
            self.playground.create_debian_environment(
                codename="bookworm", architecture=arch, variant="sbuild"
            )
            self.playground.create_debian_environment(
                codename="bookworm",
                architecture=arch,
                variant="assemblesignedsource",
            )

        root = self.orchestrate(
            task_data=MakeSignedSourceWorkflowData(
                architectures=architectures,
                binary_artifacts=LookupMultiple.model_validate(
                    [
                        f"internal@collections/name:build-{arch}"
                        for arch in (*architectures, "arm64")
                    ]
                ),
                signing_template_artifacts=LookupMultiple.model_validate(
                    [
                        {
                            "collection": "internal@collections",
                            "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                            "name__startswith": f"signing-template-{arch}-",
                        }
                        for arch in architectures
                    ]
                ),
                vendor="debian",
                codename="bookworm",
                purpose=KeyPurpose.UEFI,
                key="ABC123",
            ),
            architectures=architectures,
            create_sbuild_promises=True,
            extra_sbuild_architectures=("arm64",),
        )

        initial_sbuild_workflow = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        make_signed_source = root.children.get(
            task_type=TaskTypes.WORKFLOW, task_name="make_signed_source"
        )

        extract_for_signings = make_signed_source.children.filter(
            task_name="extractforsigning"
        )
        self.assertEqual(extract_for_signings.count(), len(architectures))

        signs = make_signed_source.children.filter(
            task_type=TaskTypes.SIGNING, task_name="sign"
        )
        self.assertEqual(signs.count(), len(architectures))

        assemble_signed_sources = make_signed_source.children.filter(
            task_name="assemblesignedsource"
        )
        self.assertEqual(assemble_signed_sources.count(), len(architectures))

        sbuild_workflows = make_signed_source.children.filter(
            task_type=TaskTypes.WORKFLOW, task_name="sbuild"
        )
        self.assertEqual(sbuild_workflows.count(), len(architectures))

        for arch in architectures:
            template_name = f"hello-{arch}-signed-template"
            extract_for_signing = extract_for_signings.get(
                task_data__input__template_artifact__startswith=(
                    f"internal@collections/name:signing-template-{arch}-"
                )
            )
            self.assertEqual(
                extract_for_signing.task_data,
                {
                    "environment": "debian/match:codename=bookworm",
                    "input": {
                        "binary_artifacts": [
                            f"internal@collections/name:build-{arch}"
                        ],
                        "template_artifact": (
                            f"internal@collections/"
                            f"name:signing-template-{arch}-{template_name}"
                        ),
                    },
                },
            )
            self.assertEqual(
                extract_for_signing.workflow_data_json,
                {
                    "display_name": (
                        f"Extract for signing ({template_name}/{arch})"
                    ),
                    "step": f"extract-for-signing-{arch}-{template_name}",
                },
            )
            self.assert_work_request_event_reactions(
                extract_for_signing,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        artifact_filters={"category": "debusine:signing-input"},
                        collection="internal@collections",
                        name_template=(
                            "extracted-for-signing-{architecture}-"
                            "{binary_package_name}"
                        ),
                        variables={
                            "$binary_package_name": "binary_package_name",
                            "architecture": arch,
                        },
                    )
                ],
            )

            unsigned = LookupMultiple.model_validate(
                {
                    "collection": "internal@collections",
                    "child_type": LookupChildType.ARTIFACT,
                    "category": ArtifactCategory.SIGNING_INPUT,
                    "name__startswith": "extracted-for-signing-",
                    "data__architecture": arch,
                }
            )
            sign = signs.get(
                task_data__unsigned=unsigned.model_dump(exclude_unset=True)
            )
            self.assertEqual(
                {
                    key: value
                    for key, value in sign.task_data.items()
                    if key != "unsigned"
                },
                {"key": "ABC123", "purpose": "uefi"},
            )
            self.assertEqual(
                sign.workflow_data_json,
                {
                    "display_name": f"Sign ({template_name}/{arch})",
                    "step": f"sign-{arch}-{template_name}",
                },
            )

            self.assert_work_request_event_reactions(
                sign,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        artifact_filters={
                            "category": "debusine:signing-output"
                        },
                        collection="internal@collections",
                        name_template=(
                            "signed-{architecture}-{template_name}-"
                            "{binary_package_name}"
                        ),
                        variables={
                            "$binary_package_name": "binary_package_name",
                            "architecture": arch,
                            "template_name": template_name,
                        },
                    )
                ],
            )

            signed = LookupMultiple.model_validate(
                {
                    "collection": "internal@collections",
                    "child_type": LookupChildType.ARTIFACT,
                    "category": ArtifactCategory.SIGNING_OUTPUT,
                    "name__startswith": f"signed-{arch}-{template_name}-",
                    "data__architecture": arch,
                }
            )
            assemble_signed_source = assemble_signed_sources.get(
                task_data__signed=signed.model_dump(exclude_unset=True)
            )

            self.assertEqual(
                {
                    key: value
                    for key, value in assemble_signed_source.task_data.items()
                    if key != "signed"
                },
                {
                    "environment": "debian/match:codename=bookworm",
                    "template": (
                        f"internal@collections/"
                        f"name:signing-template-{arch}-{template_name}"
                    ),
                },
            )
            self.assertEqual(
                assemble_signed_source.workflow_data_json,
                {
                    "display_name": (
                        f"Assemble signed source ({template_name}/{arch})"
                    ),
                    "step": f"assemble-signed-source-{arch}-{template_name}",
                },
            )
            self.assert_work_request_event_reactions(
                assemble_signed_source,
                on_success=[
                    ActionUpdateCollectionWithArtifacts(
                        artifact_filters={"category": "debian:upload"},
                        collection="internal@collections",
                        name_template=(f"signed-source-{arch}-{template_name}"),
                    )
                ],
            )

            sbuild_workflow = sbuild_workflows.get(
                task_data__input__source_artifact=(
                    f"internal@collections/"
                    f"name:signed-source-{arch}-{template_name}"
                )
            )
            self.assertEqual(
                sbuild_workflow.status, WorkRequest.Statuses.BLOCKED
            )

            self.assertEqual(
                sbuild_workflow.task_data,
                {
                    "prefix": f"signed-source-{arch}-{template_name}|",
                    "architectures": ["all", "amd64", "i386"],
                    "backend": BackendType.AUTO,
                    "input": {
                        "source_artifact": (
                            f"internal@collections/"
                            f"name:signed-source-{arch}-{template_name}"
                        ),
                        "extra_binary_artifacts": [
                            f"internal@collections/name:build-{arch}"
                        ],
                    },
                    "target_distribution": "debian:bookworm",
                },
            )
            self.assertEqual(
                sbuild_workflow.workflow_data_json,
                {
                    "display_name": f"Sbuild ({template_name}/{arch})",
                    "step": f"sbuild-{arch}-{template_name}",
                },
            )
            self.assertEqual(sbuild_workflow.event_reactions_json, {})

            # Completing each work request adds items to the internal
            # collection that can be used by the next work request.
            self.simulate_sbuild_workflow_completion(
                initial_sbuild_workflow,
                only_architecture=arch,
                binary_names=[template_name],
                use_data=False,
            )
            signing_input, _ = self.playground.create_artifact(
                category=ArtifactCategory.SIGNING_INPUT,
                data=DebusineSigningInput(binary_package_name="hello"),
                work_request=extract_for_signing,
            )
            extract_for_signing.refresh_from_db()
            self.playground.advance_work_request(
                extract_for_signing,
                result=WorkRequest.Results.SUCCESS,
                use_data=False,
            )
            self.assertEqual(
                [
                    result.artifact
                    for result in sign.lookup_multiple(
                        unsigned, expect_type=LookupChildType.ARTIFACT
                    )
                ],
                [signing_input],
            )

            signing_output, _ = self.playground.create_artifact(
                category=ArtifactCategory.SIGNING_OUTPUT,
                data=DebusineSigningOutput(
                    purpose=KeyPurpose.UEFI, fingerprint="ABC123", results=[]
                ),
                work_request=sign,
            )
            sign.refresh_from_db()
            self.playground.advance_work_request(
                sign, result=WorkRequest.Results.SUCCESS
            )
            self.assertEqual(
                [
                    result.artifact
                    for result in assemble_signed_source.lookup_multiple(
                        LookupMultiple.model_validate(
                            assemble_signed_source.task_data["signed"]
                        ),
                        expect_type=LookupChildType.ARTIFACT,
                    )
                ],
                [signing_output],
            )

            upload_artifacts = self.playground.create_upload_artifacts(
                src_name=f"hello-{arch}-signed",
                version="1.0.0",
                source=True,
                binary=False,
                workspace=assemble_signed_source.workspace,
                work_request=assemble_signed_source,
            )
            assemble_signed_source.refresh_from_db()
            self.playground.advance_work_request(
                assemble_signed_source,
                result=WorkRequest.Results.SUCCESS,
                use_data=False,
            )
            self.assertEqual(
                sbuild_workflow.lookup_single(
                    sbuild_workflow.task_data["input"]["source_artifact"],
                    expect_type=LookupChildType.ARTIFACT,
                ).artifact,
                upload_artifacts.upload,
            )

            # The sbuild sub-workflow is now marked pending, and will be
            # picked up by the scheduler.
            sbuild_workflow.refresh_from_db()
            self.assertEqual(
                sbuild_workflow.status, WorkRequest.Statuses.PENDING
            )

            # The scheduler would run on commit, but run it explicitly now
            # so that we can test what it does.
            self.schedule_and_run_workflow(root)
            sbuild_workflow.refresh_from_db()
            self.assertEqual(
                sbuild_workflow.status, WorkRequest.Statuses.RUNNING
            )

    @preserve_db_task_registry()
    def test_populate_with_artifacts(self) -> None:
        """
        Test populate: only children for specified architectures are generated.

        task_data["architectures"] specifies only one architecture.
        """
        architectures = ["amd64"]

        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                "hello", "1.0.0", "hello", "1.0.0", "amd64"
            )
        )
        upload_artifact = self.playground.create_upload_artifacts(
            src_name="coreutils", version="9.1"
        ).upload

        self.playground.create_minimal_binary_package_artifact(
            "hello2", "1.0.0", "hello2", "1.0.0", "i386"
        )
        self.create_signing_template_artifact("i386")

        root = self.orchestrate(
            task_data=MakeSignedSourceWorkflowData(
                binary_artifacts=LookupMultiple(
                    (binary_artifact.id, upload_artifact.id)
                ),
                signing_template_artifacts=LookupMultiple(
                    (self.signing_artifact.id,)
                ),
                vendor="debian",
                codename="bookworm",
                purpose=KeyPurpose.UEFI,
                key="ABC123",
                architectures=architectures,
            ),
            architectures=architectures,
            create_sbuild_promises=False,
        )

        make_signed_source = WorkRequest.objects.get(
            task_type=TaskTypes.WORKFLOW,
            task_name="make_signed_source",
            parent=root,
        )

        extract_for_signings = make_signed_source.children.filter(
            task_name="extractforsigning"
        )
        self.assertEqual(extract_for_signings.count(), len(architectures))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for tabular representation of querysets."""

import abc

from debusine.db.models import Worker
from debusine.test.django import TestCase
from debusine.web.views.base import ListViewBase
from debusine.web.views.table import TableMixin
from debusine.web.views.tables import WorkerTable
from debusine.web.views.tests.utils import ViewTestMixin


class TableMixinTests(ViewTestMixin, TestCase, abc.ABC):
    """Tests for :py:class:`TableMixin`."""

    def test_get_paginator(self) -> None:
        """Test get_paginator."""
        self.playground.create_worker()

        class TestView(TableMixin[Worker], ListViewBase[Worker]):
            model = Worker
            table_class = WorkerTable
            paginate_by = 10

        view = self.instantiate_view_class(TestView, "/")
        view.object_list = Worker.objects.all()
        ctx = view.get_context_data()
        paginator = ctx["paginator"]
        self.assertIsInstance(paginator.table, WorkerTable)
        self.assertEqual(paginator.per_page, 10)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Debusine command line interface."""

import argparse
import signal
import types
from typing import NoReturn

import argcomplete

from debusine.client.commands import Command
from debusine.client.config import ConfigHandler
from debusine.utils import debusine_version


class Cli:
    """
    Entry point for the command line debusine client.

    Usage:
        main = Cli(sys.argv[1:]) # [1:] to exclude the script name
        main.execute()
    """

    def __init__(self, argv: list[str]) -> None:
        """Initialize object."""
        self._argv = argv

    @staticmethod
    def _exit(signum: int, frame: types.FrameType | None) -> NoReturn:
        signum, frame  # fake usage for vulture
        raise SystemExit(0)

    @classmethod
    def build_argument_parser(cls) -> argparse.ArgumentParser:
        """Build the argument parser."""
        parser = argparse.ArgumentParser(
            prog='debusine',
            description='Interacts with a Debusine server.',
            formatter_class=argparse.ArgumentDefaultsHelpFormatter,
        )
        parser.add_argument(
            "--version",
            action="version",
            version=f"%(prog)s {debusine_version() or 'uninstalled-version'}",
        )

        parser.add_argument(
            '--server',
            help=(
                'Set server to be used, either by section name or as '
                'FQDN/scope (use configuration file default if not specified)'
            ),
        )

        parser.add_argument(
            '--scope',
            help=(
                'Set scope to be used (use configuration file default '
                'if not specified)'
            ),
        )

        parser.add_argument(
            '--config-file',
            default=ConfigHandler.DEFAULT_CONFIG_FILE_PATH,
            help='Config file path',
        )

        parser.add_argument("-s", "--silent", action="store_true")

        parser.add_argument(
            "-d", "--debug", action="store_true", help='Print HTTP traffic'
        )

        subparsers = parser.add_subparsers(
            help='Sub command', dest='sub-command', required=True
        )

        Command.add_subparser(
            subparsers,
            "workspace",
            argument_help="Workspace management",
            subparser_help="Workspace management subcommands",
        )
        collection_parser = subparsers.add_parser(
            "collection", help="Collection management"
        )
        collection_subparsers = collection_parser.add_subparsers(
            help="Collection management subcommands", required=True
        )
        Command.add_parsers(collection_subparsers, "collection")

        Command.add_subparser(
            collection_subparsers,
            "collection.relation",
            argument="relation",
            argument_help="Collection relation management",
            subparser_help="Collection relation management subcommands",
        )

        Command.add_subparser(
            subparsers,
            "workflow-template",
            argument_help="Workflow template management",
            subparser_help="Workflow template management subcommands",
        )
        Command.add_subparser(
            subparsers,
            "workflow",
            argument_help="Workflow management",
            subparser_help="Workflow management subcommands",
        )
        Command.add_subparser(
            subparsers,
            "work-request",
            argument_help="Work request management",
            subparser_help="Work request management subcommands",
        )
        Command.add_subparser(
            subparsers,
            "artifact",
            argument_help="Artifact management",
            subparser_help="Artifact management subcommands",
        )
        Command.add_subparser(
            subparsers,
            "asset",
            argument_help="Asset management",
            subparser_help="Asset management subcommands",
        )
        Command.add_subparser(
            subparsers,
            "task-config",
            argument_help="Task configuration management",
            subparser_help="Task configuration management subcommands",
        )
        archive_parser = subparsers.add_parser(
            "archive", help="Archive management"
        )
        archive_subparsers = archive_parser.add_subparsers(
            help="Archive management subcommands", required=True
        )
        Command.add_subparser(
            archive_subparsers,
            "archive.suite",
            argument="suite",
            argument_help="Suite management",
            subparser_help="Suite management subcommands",
        )
        Command.add_subparser(
            archive_subparsers,
            "archive.token",
            argument="token",
            argument_help="Token management",
            subparser_help="Token management subcommands",
        )

        # Add at the end so that the "object verb" format command groups show
        # up first in the help output
        Command.add_parsers(subparsers)

        argcomplete.autocomplete(parser)
        return parser

    def _parse_args(self) -> None:
        """Parse argv and store results in self.args."""
        self.parser = self.build_argument_parser()
        self.args = self.parser.parse_args(self._argv)

    def get_command(self) -> Command:
        """Instantiate the Command selected by the user."""
        self._parse_args()

        if command := Command.create(self.parser, self.args):
            return command

        sub_command = getattr(self.args, 'sub-command')
        raise AssertionError(f"Unexpected sub-command name: {sub_command}")

    def execute(self) -> None:
        """Execute the command requested by the user."""
        signal.signal(signal.SIGINT, self._exit)
        signal.signal(signal.SIGTERM, self._exit)

        command = self.get_command()
        command.run_porcelain()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for collection relation views."""

import json
import re
from typing import Any, override

from django.contrib.messages import get_messages
from django.test import RequestFactory
from django.urls import reverse
from django.utils.html import format_html
from lxml.objectify import ObjectifiedElement
from rest_framework import status

from debusine.artifacts.models import CollectionCategory
from debusine.db.models import Collection, CollectionRelation, User, Workspace
from debusine.db.playground import scenarios
from debusine.test.django import AllowAll, TestCase, override_permission
from debusine.web.views.collection_relations import (
    CollectionRelationCreateView,
    CollectionRelationDeleteView,
    CollectionRelationListView,
    CollectionRelationReorderRequiresView,
    _get_allowed_targets_by_type,
)
from debusine.web.views.tests.utils import ViewTestMixin


class CollectionRelationTestsBase(ViewTestMixin, TestCase):
    """Base class for CollectionRelation view tests."""

    trixie: Collection
    personal: Collection
    user: User

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.trixie = cls.playground.create_collection(
            name="trixie", category=CollectionCategory.SUITE
        )
        cls.personal = cls.playground.create_collection(
            name="personal", category=CollectionCategory.SUITE
        )
        cls.user = cls.playground.create_user("foo")

    def grant_user_can_create_relations(
        self, user: User, collection: Collection
    ) -> None:
        owners = self.playground.create_group_role(
            collection.workspace, Workspace.Roles.OWNER
        )
        self.playground.add_user(owners, user)

    def assert_collection_relation_information(
        self,
        modal: ObjectifiedElement,  # noqa: ARG002
        collection_relation: CollectionRelation,  # noqa: ARG002
    ) -> None:
        table = self.assertHasElement(modal, ".//table")
        rows = table.xpath(".//tbody/tr")

        self.assertEqual(len(rows), 4)

        created_at_td = rows[0].xpath("./td")[0]
        self.assertTextContentEqual(
            created_at_td,
            collection_relation.created_at.strftime("%Y-%m-%d %H:%M"),
        )

        created_by_td = rows[1].xpath("./td")[0]
        self.assertTextContentEqual(
            created_by_td, str(collection_relation.created_by_user)
        )

        updated_at_td = rows[2].xpath("./td")[0]
        self.assertTextContentEqual(
            updated_at_td, str(collection_relation.updated_at or "")
        )

        updated_by_td = rows[3].xpath("./td")[0]
        self.assertTextContentEqual(
            updated_by_td, str(collection_relation.updated_by_user or "")
        )

    def assert_table_rows(
        self,
        table: ObjectifiedElement,
        expected_collection_relations: list[CollectionRelation],
        *,
        sorting_table: bool = False,
        can_delete: bool = True,
    ) -> None:
        actual_rows = table.xpath("./tbody/tr")

        self.assertEqual(len(actual_rows), len(expected_collection_relations))

        for row, collection_relation in zip(
            actual_rows, expected_collection_relations
        ):
            tds = row.xpath("./td")

            if sorting_table:
                # First column in a sorting table is the handle
                self.assertHTMLContentsEquivalent(
                    tds[0],
                    '<td class="handle" style="cursor: grab;">'
                    '<span class="bi bi-list"></span></td>',
                )
            else:
                # First column in a non-sorting table is the type of relation
                self.assertTextContentEqual(
                    tds[0],
                    CollectionRelation.Types(collection_relation.type).label,
                )

            self.assertTextContentEqual(tds[1], str(collection_relation.target))
            self.assertTextContentEqual(
                tds[2], str(collection_relation.position or "")
            )
            self.assert_collection_relation_information(
                tds[3], collection_relation
            )
            actions_html = self.normalize_node(tds[4])

            if can_delete:
                delete_url = reverse(
                    "workspaces:collections:collection_relations:delete",
                    kwargs={
                        "wname": collection_relation.source.workspace.name,
                        "ccat": collection_relation.source.category,
                        "cname": collection_relation.source.name,
                        "pk": collection_relation.pk,
                    },
                )

                self.assertHTMLEqual(
                    actions_html,
                    '<td><a class="btn btn-outline-secondary" '
                    f'href="{delete_url}" title="Delete collection relation">'
                    '<span class="bi bi-trash"></span></a></td>',
                )
            else:
                self.assertHTMLEqual(actions_html, "<td></td>")


class CollectionRelationCreateViewTests(CollectionRelationTestsBase):
    """Tests for CollectionRelationCreateView class."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_create_relation_permission_denied_anonymous(self) -> None:
        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:create",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {},
        )
        self.assertResponseHTML(response, status_code=status.HTTP_403_FORBIDDEN)

    def test_create_relation_permission_denied_wrong_user(self) -> None:
        self.client.force_login(self.user)
        # It's logged in, but self.user does not have permission to create
        # collection relations in the workspace
        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:create",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {},
            user=self.user,
        )
        self.assertResponseHTML(response, status_code=status.HTTP_403_FORBIDDEN)

    def test_create_relation_invalid_form(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        self.assertEqual(self.personal.relations.count(), 0)
        # Invalid form data: target ID does not exist
        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:create",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {"type": "suite-based-on", "target": 0},
        )

        # view is rendered
        self.assertResponseHTML(response)

        # One error in the form (invalid target)
        assert response.context_data is not None
        error = response.context_data["form"].errors["target"][0]

        # The error is displayed
        self.assertInHTML(error, response.content.decode())

        # No relation is created
        self.assertEqual(self.personal.relations.count(), 0)

    def test_create_relation_created(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        for relation_type, expected_position in (
            (CollectionRelation.Types.SUITE_BASED_ON, None),
            (
                CollectionRelation.Types.SUITE_REQUIRES,
                1,
            ),
        ):
            with self.subTest(relation_type=str(relation_type)):
                response = self.client.post(
                    reverse(
                        "workspaces:collections:collection_relations:create",
                        kwargs={
                            "wname": self.personal.workspace.name,
                            "ccat": self.personal.category,
                            "cname": self.personal.name,
                        },
                    ),
                    {"type": relation_type, "target": self.trixie.id},
                )
                self.assertRedirects(
                    response,
                    reverse(
                        "workspaces:collections:collection_relations:list",
                        kwargs={
                            "wname": self.personal.workspace.name,
                            "ccat": self.personal.category,
                            "cname": self.personal.name,
                        },
                    ),
                )

                # The new relation is correct
                new_relation = self.personal.relations.get(type=relation_type)

                self.assertEqual(new_relation.source, self.personal)
                self.assertEqual(new_relation.target, self.trixie)
                self.assertEqual(new_relation.created_by_user, self.user)
                self.assertEqual(new_relation.position, expected_position)

                self.assertIsNone(new_relation.updated_by_user)

                # Response has the message that relation is created
                messages_list = list(get_messages(response.wsgi_request))
                self.assertEqual(len(messages_list), 1)
                self.assertEqual(
                    str(messages_list[0]),
                    format_html(
                        "Relation <code>{relation}</code> created.",
                        relation=new_relation,
                    ),
                )

    def test_get_place(self) -> None:
        view = CollectionRelationCreateView()

        view.request = RequestFactory().get("/")

        view.collection = self.personal
        place = view.get_place()

        self.assertEqual(place.breadcrumb, "Create")
        self.assertEqual(
            place.title, f"Create relation for <code>{self.personal}</code>"
        )
        self.assertEqual(
            place.url,
            reverse(
                "workspaces:collections:collection_relations:create",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

    def test_get_context_data(self) -> None:
        # Set workspaces and inheritance
        current_workspace = self.playground.get_default_workspace()
        unrelated_workspace = self.playground.create_workspace(name="Unrelated")
        parent_workspace = self.playground.create_workspace(name="Parent")

        current_workspace.set_inheritance([parent_workspace])

        # Grant permissions on these workspaces so user can view collections
        self.playground.create_group_role(
            current_workspace,
            Workspace.Roles.OWNER,
            users=[self.user],
        )
        self.playground.create_group_role(
            parent_workspace,
            Workspace.Roles.OWNER,
            users=[self.user],
        )
        self.playground.create_group_role(
            unrelated_workspace,
            Workspace.Roles.OWNER,
            users=[self.user],
        )

        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        suite_1 = self.playground.create_collection(
            name="bookworm",
            category=CollectionCategory.SUITE,
        )
        suite_2 = self.playground.create_collection(
            name="forky",
            category=CollectionCategory.SUITE,
        )

        # suite_3 is in the parent workspace: it's available for selection
        # (inheritance chain)
        suite_3 = self.playground.create_collection(
            name="suite_3",
            category=CollectionCategory.SUITE,
            workspace=parent_workspace,
        )

        # suite_4 is in an unrelated workspace: it's not available for selection
        self.playground.create_collection(
            name="suite_4",
            category=CollectionCategory.SUITE,
            workspace=unrelated_workspace,
        )

        qa_results_1 = self.playground.create_collection(
            name="qa-results-1",
            category=CollectionCategory.QA_RESULTS,
        )
        qa_results_2 = self.playground.create_collection(
            name="qa-results-2",
            category=CollectionCategory.QA_RESULTS,
        )
        qa_results_3 = self.playground.create_collection(
            name="qa-results-3",
            category=CollectionCategory.QA_RESULTS,
            workspace=parent_workspace,
        )
        self.playground.create_collection(
            name="qa-results-4",
            category=CollectionCategory.QA_RESULTS,
            workspace=unrelated_workspace,
        )

        # Existing SUITE_REQUIRES relation: this target must be excluded only
        # from the "suite-requires" entry.
        CollectionRelation.objects.create(
            source=self.personal,
            target=suite_1,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.user,
            position=1,
        )

        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:create",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

        html = self.assertResponseHTML(response)
        allowed_targets_by_type = self._extract_json_script(
            response.content, "allowed-targets-by-type"
        )

        self.assertEqual(
            allowed_targets_by_type,
            {
                CollectionRelation.Types.SUITE_BASED_ON: [
                    self.trixie.id,
                    suite_1.id,
                    suite_2.id,
                    suite_3.id,
                ],
                CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS: [
                    qa_results_1.id,
                    qa_results_2.id,
                    qa_results_3.id,
                ],
                CollectionRelation.Types.SUITE_FORKED_FROM: [
                    self.trixie.id,
                    suite_1.id,
                    suite_2.id,
                    suite_3.id,
                ],
                CollectionRelation.Types.SUITE_REQUIRES: [
                    self.trixie.id,
                    suite_2.id,
                    suite_3.id,
                ],
                CollectionRelation.Types.SUITE_TARGETING: [
                    self.trixie.id,
                    suite_1.id,
                    suite_2.id,
                    suite_3.id,
                ],
            },
        )

        # The form is displayed
        self.assertHasElement(html, "//form[@id='form']")

    def test_get_context_data_no_form(self) -> None:
        workspace = self.playground.get_default_workspace()
        # Grant permissions on these workspaces so user can view collections
        self.playground.create_group_role(
            workspace,
            Workspace.Roles.OWNER,
            users=[self.user],
        )

        self.client.force_login(self.user)

        source = self.playground.create_collection(
            name="qa-results", category=CollectionCategory.QA_RESULTS
        )

        self.grant_user_can_create_relations(self.user, source)

        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:create",
                kwargs={
                    "wname": workspace.name,
                    "ccat": source.category,
                    "cname": source.name,
                },
            ),
        )

        html = self.assertResponseHTML(response)

        # The form is displayed
        self.assertHasElement(html, "//p[@id='cannot-create-relations']")

    def test_get_form_kwargs(self) -> None:
        source = self.playground.create_collection(
            "trixie-for-kwargs", category=CollectionCategory.SUITE
        )

        view = CollectionRelationCreateView()
        view.request = RequestFactory().get("/")
        view.kwargs = {
            "wname": source.workspace.name,
            "ccat": source.category,
            "cname": source.name,
        }

        self.playground.create_collection(
            "some-suite", category=CollectionCategory.SUITE
        )
        self.playground.create_collection(
            "some-suite-qa", category=CollectionCategory.QA_RESULTS
        )

        with override_permission(Workspace, "can_configure", AllowAll):
            kwargs = view.get_form_kwargs()

        self.assertEqual(kwargs["allowed_types"], set(CollectionRelation.Types))
        self.assertEqual(kwargs["source_collection"], source)

    @staticmethod
    def _extract_json_script(response: bytes, script_id: str) -> Any:
        """Return JSON content from a <script type="application/json"> by id."""
        script_re = re.compile(
            rf'<script[^>]*\bid="{re.escape(script_id)}"[^>]*\btype="application/json"[^>]*>'
            r'(?P<content>.*?)</script>',
            re.DOTALL,
        )

        content = response.decode()

        match = script_re.search(content)
        assert match is not None

        return json.loads(match.group("content"))


class CollectionRelationDeleteViewTests(CollectionRelationTestsBase):
    """Tests for CollectionRelationDeleteView class."""

    def test_delete_relation_permission_denied_anonymous(self) -> None:
        response = self.client.delete(
            reverse(
                "workspaces:collections:collection_relations:delete",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                    "pk": 1,
                },
            ),
        )
        self.assertResponseHTML(response, status_code=status.HTTP_403_FORBIDDEN)

    def test_delete_relation_permission_denied_wrong_user(self) -> None:
        bookworm = self.playground.create_collection(
            name="bookworm", category=CollectionCategory.SUITE
        )
        collection_relation = CollectionRelation.objects.create(
            source=self.trixie,
            target=bookworm,
            type=CollectionRelation.Types.SUITE_BASED_ON,
            created_by_user=self.user,
        )

        new_user = self.playground.create_user("new-user")
        self.client.force_login(new_user)
        # It's logged in, but new_user does not have permission to create
        # collection relations in the workspace

        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:delete",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                    "pk": collection_relation.pk,
                },
            ),
            {},
            user=new_user,
        )
        self.assertResponseHTML(response, status_code=status.HTTP_403_FORBIDDEN)

    def test_delete_not_found(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:delete",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                    "pk": 0,
                },
            ),
            {},
            user=self.user,
        )

        self.assertResponseHTML(response, status_code=status.HTTP_404_NOT_FOUND)

    def test_delete(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        collection_relation = CollectionRelation.objects.create(
            source=self.personal,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.user,
            position=1,
        )

        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:delete",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                    "pk": collection_relation.pk,
                },
            ),
            {},
            user=self.user,
        )
        self.assertRedirects(
            response,
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

        self.assertFalse(
            CollectionRelation.objects.filter(
                id=collection_relation.id
            ).exists()
        )

        # Response has the message that relation is deleted
        messages_list = list(get_messages(response.wsgi_request))
        self.assertEqual(len(messages_list), 1)
        self.assertEqual(
            str(messages_list[0]),
            format_html(
                "Relation <code>{relation}</code> deleted.",
                relation=collection_relation,
            ),
        )

    def test_get_place(self) -> None:
        view = CollectionRelationDeleteView()

        collection_relation = CollectionRelation.objects.create(
            source=self.personal,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.user,
            position=1,
        )

        view.request = RequestFactory().get("/")
        view.object = collection_relation

        place = view.get_place()

        self.assertEqual(place.breadcrumb, "Delete")
        self.assertEqual(
            place.title,
            format_html(
                "Delete relation for <code>{relation}</code>",
                relation=collection_relation,
            ),
        )
        self.assertEqual(
            place.url,
            reverse(
                "workspaces:collections:collection_relations:delete",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                    "pk": collection_relation.id,
                },
            ),
        )


class CollectionRelationListViewTests(CollectionRelationTestsBase):
    """Tests for CollectionRelationlistView class."""

    no_relations_xpath: str = '//td[@id="no-relations"]'

    def test_get_place(self) -> None:
        view = CollectionRelationListView()

        view.request = RequestFactory().get("/")

        view.collection = self.personal
        place = view.get_place()

        self.assertEqual(place.breadcrumb, "Relations")
        self.assertEqual(
            place.title,
            format_html(
                "Relations of <code>{collection}</code> collection",
                collection=self.personal,
            ),
        )
        self.assertEqual(
            place.url,
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

    def test_list_with_buttons_create_resort_requires(self) -> None:
        self.grant_user_can_create_relations(self.user, self.personal)

        create_xpath = '//a[@id="create"]'
        reorder_requires_xpath = '//a[@id="show_reorder_requires"]'

        trixie_backports = self.playground.create_collection(
            name="trixie-backports", category=CollectionCategory.SUITE
        )

        # Client not logged in
        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )
        tree = self.assertResponseHTML(response)

        # No buttons to create or resort collection relations
        self.assertFalse(tree.xpath(create_xpath))
        self.assertFalse(tree.xpath(reorder_requires_xpath))

        # Logs in client
        self.client.force_login(self.user)
        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )
        tree = self.assertResponseHTML(response)

        # Create button, but no re-sorting yet (no collection relationships
        # to resort)
        self.assertTrue(tree.xpath(create_xpath))
        self.assertFalse(tree.xpath(reorder_requires_xpath))

        # Add relationships to have the button
        collection_relation_1 = CollectionRelation.objects.create(
            source=self.personal,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=1,
            created_by_user=self.user,
        )
        collection_relation_2 = CollectionRelation.objects.create(
            source=self.personal,
            target=trixie_backports,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=2,
            created_by_user=self.user,
        )

        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )
        tree = self.assertResponseHTML(response)
        self.assertHasElement(tree, reorder_requires_xpath)
        self.assertHasElement(tree, create_xpath)

        self.assertFalse(tree.xpath(self.no_relations_xpath))

        table = self.assertHasElement(tree, '//table[@id="relations"]')
        self.assert_table_rows(
            table,
            [collection_relation_1, collection_relation_2],
            can_delete=True,
        )

    def test_list_with_filter(self) -> None:
        CollectionRelation.objects.create(
            source=self.personal,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            position=1,
            created_by_user=self.user,
        )

        trixie_backports = self.playground.create_collection(
            name="trixie-backports", category=CollectionCategory.SUITE
        )

        collection_relation_targeting = CollectionRelation.objects.create(
            source=self.personal,
            target=trixie_backports,
            type=CollectionRelation.Types.SUITE_TARGETING,
            created_by_user=self.user,
        )

        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            )
            + "?relation_type=suite-targeting",
        )

        tree = self.assertResponseHTML(response)

        relations_of_type_p = self.assertHasElement(
            tree, '//p[@id="relations-of-type"]'
        )
        self.assertTextContentEqual(
            relations_of_type_p, 'Relations of type "Targeting" only. List all.'
        )

        table = self.assertHasElement(tree, '//table[@id="relations"]')
        self.assert_table_rows(
            table, [collection_relation_targeting], can_delete=False
        )

    def test_list_with_invalid_filter(self) -> None:
        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            )
            + "?relation_type=some-label-does-not-exist",
        )

        tree = self.assertResponseHTML(response)

        relations_of_type_p = self.assertHasElement(
            tree, '//p[@id="relations-of-type"]'
        )

        self.assertTextContentEqual(
            relations_of_type_p,
            'Relations of type "some-label-does-not-exist" only. List all.',
        )

        table = self.assertHasElement(tree, '//table[@id="relations"]')
        rows = table.xpath('.//tbody/tr')

        self.assertEqual(len(rows), 1)
        self.assertTextContentEqual(
            rows[0],
            "Collection personal@debian:suite does not have any "
            "relations of type some-label-does-not-exist. "
            f"List all relations for {self.personal}",
        )

    def test_list_no_relations(self) -> None:
        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

        tree = self.assertResponseHTML(response)
        self.assertHasElement(tree, self.no_relations_xpath)


class CollectionRelationReorderRequiresViewTests(CollectionRelationTestsBase):
    """Tests for CollectionRelationReorderRequires."""

    def test_reorder_relations_permission_denied_anonymous(self) -> None:
        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {},
        )
        self.assertResponseHTML(response, status_code=status.HTTP_403_FORBIDDEN)

    def test_reorder_relations_permission_denied_wrong_user(self) -> None:
        self.client.force_login(self.user)
        # It's logged in, but self.user does not have permission to create
        # collection relations in the workspace
        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {},
            user=self.user,
        )
        self.assertResponseHTML(response, status_code=status.HTTP_403_FORBIDDEN)

    def test_get_place(self) -> None:
        view = CollectionRelationReorderRequiresView()

        view.request = RequestFactory().get("/")

        view.collection = self.personal
        place = view.get_place()

        self.assertEqual(place.breadcrumb, "Reorder")
        self.assertEqual(
            place.title,
            format_html(
                "Reorder <code>Requires</code> for <code>{collection}</code>",
                collection=self.personal,
            ),
        )
        self.assertEqual(
            place.url,
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

    def test_get_list(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        trixie_backports = self.playground.create_collection(
            name="trixie-backports", category=CollectionCategory.SUITE
        )
        collection_relation_1 = CollectionRelation.objects.create(
            source=self.personal,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.user,
            position=1,
        )

        collection_relation_2 = CollectionRelation.objects.create(
            source=self.personal,
            target=trixie_backports,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.user,
            position=2,
        )

        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

        tree = self.assertResponseHTML(response)

        table = self.assertHasElement(tree, '//table[@id="relations"]')
        self.assert_table_rows(
            table,
            [collection_relation_1, collection_relation_2],
            sorting_table=True,
            can_delete=True,
        )

    def test_post(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        trixie_backports = self.playground.create_collection(
            name="trixie-backports", category=CollectionCategory.SUITE
        )

        relation_1 = CollectionRelation.objects.create(
            source=self.personal,
            target=self.trixie,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.user,
            position=1,
        )
        relation_2 = CollectionRelation.objects.create(
            source=self.personal,
            target=trixie_backports,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=self.user,
            position=2,
        )

        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {"order": f"{relation_2.id},{relation_1.id}"},
        )

        self.assertRedirects(
            response,
            reverse(
                "workspaces:collections:collection_relations:list",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

        relation_1.refresh_from_db()
        relation_2.refresh_from_db()

        # Reordered
        self.assertEqual(relation_1.position, 2)
        self.assertEqual(relation_2.position, 1)

        messages_list = list(get_messages(response.wsgi_request))
        self.assertEqual(len(messages_list), 1)
        self.assertEqual(str(messages_list[0]), "Requires relations reordered.")

    def test_post_without_order_invalid_request(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {},
        )

        self.assertResponse400(response, 'Missing "order" POST parameter')

    def test_post_invalid_order_invalid_request(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {"order": "missing numbers"},
        )

        self.assertResponse400(response, 'Invalid "order" format')

    def test_post_ids_do_not_match_db_invalid_request(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        response = self.client.post(
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
            {"order": "1,2,3,4"},
        )

        self.assertResponse400(
            response, "Submitted ids and DB ids do not match"
        )

    def test_get_list_empty(self) -> None:
        self.client.force_login(self.user)
        self.grant_user_can_create_relations(self.user, self.personal)

        response = self.client.get(
            reverse(
                "workspaces:collections:collection_relations:reorder_requires",
                kwargs={
                    "wname": self.personal.workspace.name,
                    "ccat": self.personal.category,
                    "cname": self.personal.name,
                },
            ),
        )

        tree = self.assertResponseHTML(response)

        no_relations = self.assertHasElement(tree, '//td[@id="no-relations"]')

        self.assertTextContentEqual(
            no_relations, "No relations of type Requires."
        )


class HelperFunctionsTests(TestCase):
    """Helper functions tests."""

    scenario = scenarios.DefaultContext(set_current=True)

    collection: Collection

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()

        cls.collection = cls.playground.create_collection(
            "trixie", category=CollectionCategory.SUITE
        )

    def test_get_allowed_targets_by_type_no_collections_available(self) -> None:
        with override_permission(Workspace, "can_configure", AllowAll):
            self.assertEqual(
                _get_allowed_targets_by_type(self.collection),
                {
                    CollectionRelation.Types.SUITE_FORKED_FROM: [],
                    CollectionRelation.Types.SUITE_TARGETING: [],
                    CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS: [],
                    CollectionRelation.Types.SUITE_REQUIRES: [],
                    CollectionRelation.Types.SUITE_BASED_ON: [],
                },
            )

    def test_get_allowed_targets_by_type_collections_used(self) -> None:
        suite = self.playground.create_collection(
            "bookworm", category=CollectionCategory.SUITE
        )

        qa_results = self.playground.create_collection(
            "qa-results", category=CollectionCategory.QA_RESULTS
        )

        user = self.playground.get_default_user()
        CollectionRelation.objects.create(
            source=self.collection,
            target=suite,
            type=CollectionRelation.Types.SUITE_FORKED_FROM,
            created_by_user=user,
        )
        CollectionRelation.objects.create(
            source=self.collection,
            target=suite,
            type=CollectionRelation.Types.SUITE_TARGETING,
            created_by_user=user,
        )
        CollectionRelation.objects.create(
            source=self.collection,
            target=qa_results,
            type=CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS,
            created_by_user=user,
        )
        CollectionRelation.objects.create(
            source=self.collection,
            target=suite,
            type=CollectionRelation.Types.SUITE_REQUIRES,
            created_by_user=user,
            position=1,
        )
        CollectionRelation.objects.create(
            source=self.collection,
            target=suite,
            type=CollectionRelation.Types.SUITE_BASED_ON,
            created_by_user=user,
        )

        with override_permission(Workspace, "can_configure", AllowAll):
            self.assertEqual(
                _get_allowed_targets_by_type(self.collection),
                {
                    CollectionRelation.Types.SUITE_FORKED_FROM: [],
                    CollectionRelation.Types.SUITE_TARGETING: [],
                    CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS: [],
                    CollectionRelation.Types.SUITE_REQUIRES: [],
                    CollectionRelation.Types.SUITE_BASED_ON: [],
                },
            )

    def test_get_allowed_targets_by_type_available_collections(self) -> None:
        suite = self.playground.create_collection(
            "bookworm", category=CollectionCategory.SUITE
        )

        qa_results = self.playground.create_collection(
            "qa-results", category=CollectionCategory.QA_RESULTS
        )

        with override_permission(Workspace, "can_configure", AllowAll):
            self.assertEqual(
                _get_allowed_targets_by_type(self.collection),
                {
                    CollectionRelation.Types.SUITE_FORKED_FROM: [suite.id],
                    CollectionRelation.Types.SUITE_TARGETING: [suite.id],
                    CollectionRelation.Types.SUITE_DEFAULT_QA_RESULTS: [
                        qa_results.id
                    ],
                    CollectionRelation.Types.SUITE_REQUIRES: [suite.id],
                    CollectionRelation.Types.SUITE_BASED_ON: [suite.id],
                },
            )



#!/usr/bin/env python3

# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""
Debusine integration tests.

Test simplesystemimagebuild related code.
"""

import subprocess
import unittest
from typing import override

import yaml

from debusine.artifacts.models import ArtifactCategory
from utils.client import Client
from utils.common import Configuration
from utils.integration_test_helpers_mixin import IntegrationTestHelpersMixin
from utils.server import DebusineServer
from utils.worker import Worker


class IntegrationTaskSimpleSystemImageBuildTests(
    IntegrationTestHelpersMixin, unittest.TestCase
):
    """
    Integration test for the simplesystemimagebuild task.

    These tests assume:
    - debusine-server is running
    - debusine-worker is running (connected to the server)
    - debusine-client is correctly configured
    """

    TASK_NAME = "simplesystemimagebuild"

    @override
    def setUp(self) -> None:
        super().setUp()
        # If debusine-server or nginx was launched just before the
        # integration-tests-simplesystemimagebuild.py is launched the
        # debusine-server might not be yet available. Let's wait for the
        # debusine-server to be reachable if it's not ready yet
        self.assertTrue(
            DebusineServer.wait_for_server_ready(),
            f"debusine-server should be available (in "
            f"{Configuration.get_base_url()}) before the integration tests "
            f"are run",
        )

        self.worker = Worker()

    def test_simplesystemimagebuild(self) -> None:
        """Create a job to build a system image. Verify it built."""
        task_data = {
            "bootstrap_options": {
                "architecture": subprocess.check_output(
                    ["dpkg", "--print-architecture"], text=True
                ).strip(),
                "variant": "minbase",
            },
            "bootstrap_repositories": [
                {
                    "mirror": "http://deb.debian.org/debian",
                    "suite": "bookworm",
                    "components": ["main"],
                }
            ],
            "disk_image": {
                "filename": "debian",
                "format": "qcow2",
                # This is considerably smaller than the default.  It's
                # amd64-specific, but this test is only run on amd64 anyway.
                "kernel_package": "linux-image-cloud-amd64",
                "partitions": [
                    {
                        "size": 1,
                        "filesystem": "ext4",
                    },
                ],
            },
        }

        work_request_id = Client.execute_command(
            "work-request",
            "create",
            "--yaml",
            self.TASK_NAME,
            stdin=yaml.safe_dump(task_data),
        )["id"]

        # The worker should get the new work request and start executing it
        status = Client.wait_for_work_request_completed(
            work_request_id, "success", timeout=2400
        )
        if not status:
            self.print_work_request_debug_logs(work_request_id)
        self.assertTrue(status)

        work_request = Client.execute_command(
            "work-request", "show", work_request_id
        )

        debian_system_image_artifacts = 0
        for artifact in work_request["artifacts"]:
            if artifact["category"] == ArtifactCategory.SYSTEM_IMAGE:
                debian_system_image_artifacts += 1

                self.assertIn("debian.qcow2", artifact["files"].keys())

        self.assertEqual(debian_system_image_artifacts, 1)
